class UpdateEvent(object):
    ACCESS_TOKEN_SUCCESS = "access_token_success"
    UPDATE_ACCESS_TOKEN = "update_access_token"
    ACCESS_TOKEN_UNAUTHORIZED = "access_token_unauthorized"
    ACCESS_TOKEN_FAILED = "access_token_failed"
    RECONNECTING = "access_token_failed"
    CLOSE_SESSION = "close_session"
    REFRESH_TOKEN_SUCCESS = "refresh_token_success"
    REFRESH_TOKEN_FAILED = "refresh_token_failed"
    REFRESH_TOKEN_BAD = "refresh_token_bad"
    REFRESH_TOKEN_EXPIRED = "refresh_token_expired"
    AUTHENTICATION_SUCCESS = "authentication_success"
    AUTHENTICATION_FAILED = "authentication_failed"



# coding: utf8
# contract_gen 2020-06-16 10:26:10.693191

__all__ = ["Format"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class Format(Enum):
    """
    The enumerate that specifies whether the calculated volatilities
    """

    LIST = "List"
    MATRIX = "Matrix"
    N_DIMENSIONAL_ARRAY = "NDimensionalArray"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(Format, _FORMAT_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_FORMAT_VALUES_IN_LOWER_BY_FORMAT, some)


_FORMAT_VALUES = tuple(t.value for t in Format)
_FORMAT_VALUES_IN_LOWER_BY_FORMAT = {
    name.lower(): item for name, item in Format.__members__.items()
}



import abc
import math
import threading
from typing import TYPE_CHECKING, Callable, Optional, Union

from pandas import NA

from .._tools import RepeatedTimer

if TYPE_CHECKING:
    from ._stream_update_handler import (
        CollectUpdates_StreamUpdateHandler,
        BuildDF_StreamUpdateHandler,
        StreamUpdateHandler,
    )
    from ._ohlc_builder import OHLCBuilder


class RecordMixin:
    logger = None
    update_handler: "StreamUpdateHandler" = None
    ohlc_builder: "OHLCBuilder" = None
    on_record_callback: Callable = None

    def check_recorded_updates(self):
        for (
            steam_name,
            updates,
        ) in self.update_handler.updates_by_stream_name.items():
            if updates:
                self.record()
                break

    def record(self):
        updates_by_stream_name = self.update_handler.updates_by_stream_name
        self.ohlc_builder.build(updates_by_stream_name)
        ohlc_df = self.ohlc_builder._last_recorded_ohlc_updates

        if self.on_record_callback:
            try:
                if not ohlc_df.empty:
                    ohlc_df.fillna(NA, inplace=True)

                t = threading.Thread(
                    target=self.on_record_callback, args=(ohlc_df, self._recorder)
                )
                t.start()

            except Exception as error:
                self.logger.exception("Error occurred in user's callback function.")
                self.logger.exception(error)


class RecordingControl(abc.ABC):
    @abc.abstractmethod
    def start_recording(self):
        pass

    @abc.abstractmethod
    def stop_recording(self):
        pass

    @abc.abstractmethod
    def delete_recording(self):
        pass


class NoBlocking_RecordingControl(RecordingControl):
    def __init__(
        self,
        update_handler: Union[
            "CollectUpdates_StreamUpdateHandler", "BuildDF_StreamUpdateHandler"
        ],
    ) -> None:
        self.update_handler = update_handler

    def start_recording(self):
        self.update_handler.start_handling()

    def stop_recording(self):
        if self.update_handler._ticks_per_bar != 1:
            self.update_handler.check_recorded_updates()

        self.update_handler.stop_handling()

    def delete_recording(self):
        self.update_handler.dispose()


class Blocking_RecordingControl(RecordingControl):
    def __init__(
        self,
        update_handler: Union[
            "CollectUpdates_StreamUpdateHandler", "BuildDF_StreamUpdateHandler"
        ],
    ) -> None:
        self.update_handler = update_handler
        self.blocking = threading.Event()

    def start_recording(self, interval: int = 0):
        self.update_handler.start_handling()
        self.blocking.wait(interval)

    def stop_recording(self):
        if self.update_handler._ticks_per_bar != 1:
            self.update_handler.check_recorded_updates()
        self.update_handler.stop_handling()
        self.blocking.clear()

    def delete_recording(self):
        self.update_handler.dispose()


class RepeatNonBlocking_RecordingControl(RecordingControl, RecordMixin):
    timer: Optional[RepeatedTimer] = None

    def __init__(
        self,
        on_record_callback: Callable,
        update_handler: "CollectUpdates_StreamUpdateHandler",
        ohlc_builder: "OHLCBuilder",
        logger,
        recorder,
    ) -> None:
        self.logger = logger
        self.ohlc_builder = ohlc_builder
        self.update_handler = update_handler
        self.on_record_callback = on_record_callback
        self._recorder = recorder

    def start_recording(self, interval: int):
        self.update_handler.start_handling()
        self.timer = RepeatedTimer(function=self.record, interval=interval)
        self.timer.start()

    def stop_recording(self):
        self.check_recorded_updates()
        self.update_handler.stop_handling()
        self.ohlc_builder.dataframes = []
        self.timer.cancel()

    def delete_recording(self):
        self.ohlc_builder.dispose()
        self.update_handler.dispose()


class RepeatBlocking_RecordingControl(RecordingControl, RecordMixin):
    def __init__(
        self,
        duration: int,
        on_record_callback: Callable,
        update_handler: "CollectUpdates_StreamUpdateHandler",
        ohlc_builder: "OHLCBuilder",
        logger,
        recorder: "PricingRecorder",
    ) -> None:
        self.logger = logger
        self.ohlc_builder = ohlc_builder
        self.update_handler = update_handler
        self.on_record_callback = on_record_callback
        self.duration = duration
        self.blocking = threading.Event()
        self._recorder = recorder

    def start_recording(self, interval: int):
        self.update_handler.start_handling()

        remainder, count = math.modf(self.duration / interval)
        count = int(count)
        for _ in range(count):
            self.blocking.wait(interval)
            self.record()

        if remainder:
            remainder = self.duration - (count * interval)
            self.blocking.wait(remainder)
            self.record()

    def stop_recording(self):
        self.update_handler.stop_handling()
        self.ohlc_builder.dataframes = []
        self.blocking.clear()

    def delete_recording(self):
        self.ohlc_builder.dispose()
        self.update_handler.dispose()



class LogReporter:
    """
    The class for convenient use log's methods in child classes.

    Methods
    -------
    log(level, msg, *args, **kwargs)
        Log 'msg % args' with the integer severity 'level'.
    warning(msg, *args, **kwargs)
        Log 'msg % args' with severity 'WARNING'.
    error(msg, *args, **kwargs)
        Log 'msg % args' with severity 'ERROR'.
    debug(msg, *args, **kwargs)
        Log 'msg % args' with severity 'DEBUG'.
    info(msg, *args, **kwargs)
        Log 'msg % args' with severity 'INFO'.
    """

    def __init__(self, logger):
        self.log = logger.log
        self.warning = logger.warning
        self.error = logger.error
        self.debug = logger.debug
        self.info = logger.info


class _LogReporter:
    """
    The class for convenient use log's methods in child classes.

    Methods
    -------
    _log(level, msg, *args, **kwargs)
        Log 'msg % args' with the integer severity 'level'.
    _warning(msg, *args, **kwargs)
        Log 'msg % args' with severity 'WARNING'.
    _error(msg, *args, **kwargs)
        Log 'msg % args' with severity 'ERROR'.
    _debug(msg, *args, **kwargs)
        Log 'msg % args' with severity 'DEBUG'.
    _info(msg, *args, **kwargs)
        Log 'msg % args' with severity 'INFO'.
    """

    def __init__(self, logger):
        self._log = logger.log
        self._warning = logger.warning
        self._error = logger.error
        self._debug = logger.debug
        self._info = logger.info



# coding: utf-8
import enum
import re
import traceback
from collections.abc import Callable
from datetime import timedelta
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd

from ._stream_cache import TradeDataStreamCache
from .._content_type import ContentType
from ..._core.session import get_valid_session
from ...delivery._stream import StreamStateEvent, StreamStateManager
from ...delivery._stream._stream_factory import create_rdp_stream
from ..._tools import tds_datetime_adapter, cached_property

if TYPE_CHECKING:
    from ...delivery._stream import _RDPStream


def convert_start_datetime(start_datetime):
    start_datetime_object = tds_datetime_adapter.convert(start_datetime)
    if start_datetime_object is not None:
        start_datetime_object = start_datetime_object + timedelta(microseconds=-1)
    if start_datetime_object is not None:
        start_datetime = tds_datetime_adapter.get_str(start_datetime_object)
    else:
        start_datetime = None
    return start_datetime


class Events(enum.Enum):
    """Events"""

    No = "None"
    Full = "Full"


class FinalizedOrders(enum.Enum):
    """Finalized order in cached"""

    No = "None"
    P1D = "P1D"


class UniverseTypes(enum.Enum):
    """Universe Types"""

    RIC = "RIC"
    Symbol = "Symbol"
    UserID = "UserID"


class TradeDataStream(TradeDataStreamCache, StreamStateManager):
    """
    Open a streaming trading analytics subscription.

    Parameters
    ----------
    universe: list
        a list of RIC or symbol or user's id for retrieving trading analytics data.

    fields: list
        a list of enumerate fields.
        Default: None

    universe_type: enum
        a type of given universe can be RIC, Symbol or UserID.
        Default: UniverseTypes.RIC

    finalized_orders: bool
        enable/disable the cached of finalized order of current day in the streaming.
        Default: False

    filters: list
        set the condition of subset of trading streaming data
        Default: None

    on_add: callable object (stream, add_message)
        Called when the stream on summary order of  universe is added by the server.
        This callback is called with the reference to the stream object and the universe
        full image.
        Default: None

    on_update: callable object (stream, update_message)
        Called when the stream on summary order of universe is updated by the server.
        (can be partial or full)
        This callback is called with the reference to the stream object and
        the universe update.
        Default: None

    on_remove: callable object (stream, remove_message)
        Called when the stream on summary order of universe is removed by the server.
        This callback is called with the reference to the stream object and
        the universe removed.
        Default: None

    on_event: callable object (stream, event_message)
        Called when the stream on order of universe has new event by the server.
        This callback is called with the reference to the stream object and
        the universe has a order event.
        Default: None

    on_state: callable object (stream, state_message):
        Called when the stream has new state.
        Default: None

    on_complete: callable object (stream):
        Called when the stream has complete receiving a cache data.
        Default: None

    """

    def __init__(
        self,
        universe: list = None,
        fields: list = None,
        universe_type=None,
        events: object = None,
        finalized_orders: object = None,
        filters: list = None,
        extended_params: dict = None,
        session=None,
        api: str = None,
        on_add: Callable = None,
        on_update: Callable = None,
        on_remove: Callable = None,
        on_event: Callable = None,
        on_state: Callable = None,
        on_complete: Callable = None,
        max_order_summary_cache: int = None,
        max_order_events_cache: int = None,
    ):
        self._session = get_valid_session(session)

        if universe_type is None:
            universe_type = UniverseTypes.UserID

        if events is None:
            events = Events.No

        if finalized_orders is None:
            finalized_orders = FinalizedOrders.No

        #   validate arguments
        if not isinstance(universe_type, UniverseTypes):
            #   invalid universe type
            raise TypeError(f"ERROR!!! Invalid universe type[{universe_type}].")

        #   cache
        #       order summary
        if max_order_summary_cache is None:
            max_order_summary_cache = 1500
        #       order events
        if max_order_events_cache is None:
            max_order_events_cache = 100

        self._universe = universe
        self._fields = fields
        self._universe_type = universe_type
        self._event_details = events
        self._finalized_orders = finalized_orders
        self._filters = filters
        self._extended_params = extended_params

        #   initialize the parent classes
        TradeDataStreamCache.__init__(
            self, self._session, max_order_summary_cache, max_order_events_cache
        )
        StreamStateManager.__init__(self, logger=self._session.logger())

        #   callback functions
        #       order summary
        self._on_add_cb = on_add
        self._on_update_cb = on_update
        self._on_remove_cb = on_remove
        #       order events
        self._on_event_cb = on_event

        #       state
        self._on_state_cb = on_state

        #       complete
        self._on_complete_cb = on_complete

        #   order summary header field names
        #       dictionary of header information
        self._order_summary_headers = None
        #       list of header names
        self._order_summary_header_names = None

        #   completed message
        self._is_completed = False

    @cached_property
    def _stream(self) -> "_RDPStream":
        stream = create_rdp_stream(
            ContentType.STREAMING_TRADING,
            session=self._session,
            universe=self._universe,
            view=[i for i in self._fields] if self._fields is not None else None,
            parameters=self._parameters,
            extended_params=self._extended_params,
        )
        stream.on_response(self.__on_response)
        stream.on_update(self.__on_update)
        stream.on_alarm(self.__on_alarm)
        stream.on(StreamStateEvent.CLOSED, self.close)
        return stream

    @property
    def _parameters(self):
        #   build RDP item stream parameters
        parameters = {
            "universeType": self._universe_type.value,
            "events": self._event_details.value,
            "finalizedOrders": self._finalized_orders.value,
        }
        if self._filters is not None:
            parameters["filters"] = self._filters

        return parameters

    def _do_open(self, *args, **kwargs):
        self._clear_caches()

        self._session.debug(
            f"Start asynchronously StreamingTDS subscription {self._stream.id} for {self._universe}"
        )

        self._stream.open()

    def _do_close(self, *args, **kwargs):
        self._session.debug(
            f"Stop asynchronously StreamingTDS subscription {self._stream.id} for {self._universe}"
        )
        self._stream.close()

    def get_order_summary(
        self,
        universe: list = None,
        start_datetime: object = None,
        end_datetime: object = None,
    ):
        data = np.array(
            [order_summary for order_summary in self._order_summary_dict.values()]
        )
        df = pd.DataFrame.from_records(data, columns=self._order_summary_header_names)

        #   check for filter
        start_datetime = convert_start_datetime(start_datetime)
        end_datetime = (
            tds_datetime_adapter.get_str(end_datetime) if end_datetime else None
        )
        if (universe and "RIC" in df.columns) or (
            (start_datetime or end_datetime) and ("OrderTime" in df.columns)
        ):
            #   do filter by universe and/or order time

            if universe:
                #   filter by universe
                df = df[df.RIC.isin(universe)]

            if start_datetime:
                #   filter by start date/time,
                #   add for a roundoff error when compare datetime by a microsecond
                df = df[df.OrderTime >= start_datetime]

            if end_datetime:
                #   filter by end date/time
                df = df[df.OrderTime <= end_datetime]

            #   done
            return df

        else:
            #   no filter required
            return df

    def get_order_events(
        self,
        universe: list = None,
        start_datetime: object = None,
        end_datetime: object = None,
    ):
        """get the snapshot of order events in dataframe"""
        data = np.array(self.get_last_order_events())
        start_datetime = convert_start_datetime(start_datetime)
        end_datetime = (
            tds_datetime_adapter.get_str(end_datetime) if end_datetime else None
        )

        #   check empty case
        if data.size == 0:
            #   no data, so return None
            return None

        #   build column name
        order_event_column_name = [
            "OrderKey",
        ]
        order_event_column_name.extend(self._event_column_names)

        #   get row and column for data
        data_num_rows, data_num_cols = data.shape
        assert data_num_cols == len(order_event_column_name)

        #   build dataframe
        df = pd.DataFrame.from_records(data, columns=order_event_column_name)

        #   check for filter
        if (universe and "OrderKey" in df.columns) or (
            (start_datetime or end_datetime) and ("EventTime" in df.columns)
        ):
            #   do filter by universe

            if universe:
                #   filter by universe
                df = df[df.OrderKey.str.contains("|".join(universe), regex=True)]

            if start_datetime:
                #   filter by start date/time,
                #   add for a roundoff error when compare datetime by a microsecond
                df = df[df.EventTime >= start_datetime]

            if end_datetime:
                #   filter by end date/time
                df = df[df.EventTime <= end_datetime]

            #   done
            return df

        else:
            #   no filter required
            return df

    ####################################################
    #   callbacks

    #   RDP
    def __on_response(self, stream: object, response: dict):
        """extract the response order summaries, order events and state"""

        #   extract headers, data, message and state
        #       headers
        assert "headers" in response
        self._order_summary_headers = response.get("headers", None)
        assert self._order_summary_headers is not None
        assert isinstance(self._order_summary_headers, list)

        #   generate a list of header name
        self._order_summary_header_names = [
            item.get("id", None) for item in self._order_summary_headers
        ]

        #       data
        if "data" in response:
            #   extract order summaries
            added_order_summary_data_list = response.get("data", None)
            assert added_order_summary_data_list is not None
            assert isinstance(added_order_summary_data_list, list)

            #   call the callback function on_add

            #   loop over all new order summary data
            for added_order_summary_data in added_order_summary_data_list:
                #   call the on_add callback function
                self._on_add(stream, added_order_summary_data)

        #       events
        if "messages" in response:
            #   extract order events
            order_event_list = response.get("messages", None)
            assert order_event_list is not None
            assert isinstance(order_event_list, list)

            #   call the callback function on_event

            #   loop over all new order events
            for order_event in order_event_list:
                #   call the on_event callback function
                self._on_event(stream, order_event)

        #       state
        if "state" in response:
            #   extract state
            state = response.get("state", None)
            assert isinstance(state, dict)

            #   call on_state callback
            self._on_state(self, state)

    def __on_update(self, stream, update):
        """extract the update (add/update/remove)
        order summaries and new order status."""

        #   extract data(add), update, remove and messages (order events)
        #       data
        if "data" in update:
            #   update contains a new order summary data
            added_order_summary_data_list = update.get("data", None)
            assert added_order_summary_data_list is not None
            assert isinstance(added_order_summary_data_list, list)

            #   loop over all new order summary data
            for added_order_summary_data in added_order_summary_data_list:
                #   call the on_add callback function
                self._on_add(stream, added_order_summary_data)

        #       update
        if "update" in update:
            #   update contains an updated order summary data
            updated_order_summary_dict_list = update.get("update", None)
            assert updated_order_summary_dict_list is not None
            assert isinstance(updated_order_summary_dict_list, list)

            #   loop over all updated order summary data
            for updated_order_summary_dict in updated_order_summary_dict_list:
                #   call the on_update callback function
                self._on_update(stream, updated_order_summary_dict)

        #       remove
        if "remove" in update:
            #   update contains a removed order summary data
            removed_order_summary_data_list = update.get("remove", None)
            assert removed_order_summary_data_list is not None
            assert isinstance(removed_order_summary_data_list, list)

            #   loop over all removed order summary data
            for removed_order_summary_data in removed_order_summary_data_list:
                #   call the on_update callback function
                self._on_remove(stream, removed_order_summary_data)

        #       messages
        if "messages" in update:
            #   update contains a order event.
            order_event_list = update.get("messages", None)
            assert order_event_list is not None
            assert isinstance(order_event_list, list)

            #   loop over all new order events
            for order_event in order_event_list:
                #   call the on_event callback function
                self._on_event(stream, order_event)

            #   update the cache
            self._add_order_event(order_event)

        #       state
        if "state" in update:
            #   extract state
            state = update.get("state", None)
            assert isinstance(state, dict)

            #   call on_state callback
            self._on_state(self, state)

    def __on_alarm(self, stream, alarm):
        pass

    #   TDS
    def _on_add(self, stream, add_message: list):
        """order summary add callback from RDPStream"""
        assert len(add_message) > 1

        #   update the cache
        order_key = add_message[0]
        self._add_order_summary(order_key, add_message)

        #   call the add callback function
        if self._on_add_cb:
            #   convert add message to a dict between field name and value
            add_message_dict = dict(zip(self._order_summary_header_names, add_message))
            #   valid on_add callback
            try:
                self._on_add_cb(self, add_message_dict)
            except Exception as e:
                #   on_add callback has an exception
                self._session.error(
                    f"StreamingTDS on_add callback raised exception: {e!r}"
                )
                self._session.debug(f"{traceback.format_exc()}")

    def _on_update(self, stream, update_message: dict):
        """order summary update callback from RDPStream"""

        #   update the cache
        assert "key" in update_message
        order_key = update_message["key"]
        self._update_order_summary(order_key, update_message)

        #   call the update callback function
        if self._on_update_cb:
            #   valid on_update callback
            try:
                self._on_update_cb(self, update_message)
            except Exception as e:
                #   on_update callback has an exception
                self._session.error(
                    f"StreamingTDS on_update callback raised exception: {e!r}"
                )
                self._session.debug(f"{traceback.format_exc()}")

    def _on_remove(self, stream, remove_message: dict):
        """order summary remove callback from RDPStream"""
        #   call the remove callback function
        if self._on_remove_cb:
            #   valid on_remove callback
            try:
                self._on_remove_cb(self, remove_message)
            except Exception as e:
                #   on_remove callback has an exception
                self._session.error(
                    f"StreamingTDS on_remove callback raised exception: {e!r}"
                )
                self._session.debug(f"{traceback.format_exc()}")

    def _on_event(self, stream, event_message: dict):
        """order event callback from RDPStream"""
        #   call the event callback function
        if self._on_event_cb:
            #   valid on_event callback
            try:
                self._on_event_cb(self, event_message)
            except Exception as e:
                #   on_event callback has an exception
                self._session.error(
                    f"StreamingTDS on_event callback raised exception: {e!r}"
                )
                self._session.debug(f"{traceback.format_exc()}")

    def _on_state(self, stream, state_message: dict):
        """state event callback"""

        # warning TEMPORARY SOLUTION HANDLE on_complete
        if "message" in state_message:
            #   handle the queueSize in state message
            message = state_message.get("message", None)
            assert message is not None

            #   use regular expression to determine for a queue size message or not?
            re_matched = re.match(r"^queueSize=(?P<queue_size>[0-9]+)", message)
            if re_matched is not None:
                #   this is a queue size message, so check for completion of cache
                re_matched_group_dict = re_matched.groupdict()
                assert "queue_size" in re_matched_group_dict
                queue_size_str = re_matched_group_dict.get("queue_size", None)

                assert queue_size_str is not None
                assert isinstance(queue_size_str, str)

                #   cast to int and check it
                queue_size = int(queue_size_str)
                if queue_size == 0:
                    #   cache is empty now, so call complete callback
                    self._on_complete(self)

        #   call the state callback function
        if self._on_state_cb:
            #   valid on_state callback
            try:
                self._on_state_cb(self, state_message)
            except Exception as e:
                #   on_state callback has an exception
                self._session.error(
                    f"StreamingTDS on_state callback raised exception: {e!r}"
                )
                self._session.debug(f"{traceback.format_exc()}")

    def _on_complete(self, stream):
        """complete event callback"""
        #   this function only call once
        if self._is_completed:
            #   already call the on_complete callback
            return

        #   call the complete callback function
        if self._on_complete_cb:
            #   valid _on_complete callback
            try:
                self._on_complete_cb(self)
            except Exception as e:
                #   on_complete callback has an exception
                self._session.error(
                    f"StreamingTDS _on_complete callback raised exception: {e!r}"
                )
                self._session.debug(f"{traceback.format_exc()}")

        #   set the completed flags that it won't call this callback again.
        self._is_completed = True



# coding: utf-8

from .grant import Grant


class GrantRefreshToken(Grant):
    def __init__(self, *args, **kwargs):
        super(GrantRefreshToken, self).__init__(*args, **kwargs)
        self._refresh_token = kwargs.get("refresh_token")
        self._token_scope = kwargs.get("token_scope", "trapi")

    def get_refresh_token(self):
        return self._refresh_token

    def refresh_token(self, refresh_token):
        self._refresh_token = refresh_token
        return self

    def get_token_scope(self):
        return self._token_scope

    def with_scope(self, token_scope):
        if token_scope:
            self._token_scope = token_scope
        return self



# coding: utf8

from typing import Optional

from .._instrument_pricing_parameters import InstrumentPricingParameters
from ._enums import (
    FxSwapCalculationMethod,
    OptionVolatilityType,
    PriceSide,
    PricingModelType,
    TimeStamp,
    VolatilityModel,
)
from ._models import (
    BidAskMid,
    InterpolationWeight,
    PayoutScaling,
)


class PricingParameters(InstrumentPricingParameters):
    """
    Parameters
    ----------
    atm_volatility_object : BidAskMid, optional
        At the money volatility at Expiry
    butterfly10_d_object : BidAskMid, optional
        BF 10 Days at Expiry
    butterfly25_d_object : BidAskMid, optional
        BF 25 Days at Expiry
    domestic_deposit_rate_percent_object : BidAskMid, optional
        Domestic Deposit Rate at Expiry
    foreign_deposit_rate_percent_object : BidAskMid, optional
        Foreign Deposit Rate at Expiry
    forward_points_object : BidAskMid, optional
        Forward Points at Expiry
    fx_spot_object : BidAskMid, optional
        Spot Price
    fx_swap_calculation_method : FxSwapCalculationMethod, optional
        The method we chose to price outrights using or not implied deposits. Possible
        values are:   FxSwap (compute outrights using swap points),
        FxSwapImpliedFromDeposit (compute swap points using deposits of currency1 and
        currency2),   DepositCcy1ImpliedFromFxSwap (compute currency1 deposits using
        swap points),   DepositCcy2ImpliedFromFxSwap (compute currency2 deposits using
        swap points).   Optional. Defaults to 'DepositCcy2ImpliedFromFxSwap'.
    implied_volatility_object : BidAskMid, optional
        Implied volatility at Expiry
    interpolation_weight : InterpolationWeight, optional
        Vol Term Structure Interpolation
    option_price_side : PriceSide, optional
        Quoted option's price side to use for pricing Analysis: Bid(Bid value), Ask(Ask
        value), Mid(Mid value), Last(Last value). Optional. By default the "Mid" price
        of the option is used.
    option_time_stamp : TimeStamp, optional
        Define how the timestamp of the option is selected:
        - Open: the opening value of the valuation_date or if not available the close of
          the previous day is used.
        - Default: the latest snapshot is used when valuation_date is today, the close
          price when valuation_date is in the past.
    payout_custom_dates : string, optional
        Define the custom dates for the Payout/Volatility chart
    payout_scaling_interval : PayoutScaling, optional
        Define the range for the Payout/Volatility chart
    price_side : PriceSide, optional
        The enumerate that specifies whether bid, ask or mid is used to price the Fx
        option. Possible values:
        - Bid
        - Ask
        - Mid
        - Last
    pricing_model_type : PricingModelType, optional
        The enumeration that specifies the model type of pricing. Possible values:
        - BlackScholes
        - Whaley
        - Binomial
        - Trinomial
        - VannaVolga
    risk_reversal10_d_object : BidAskMid, optional
        RR 10 Days at Expiry
    risk_reversal25_d_object : BidAskMid, optional
        RR 25 Days at Expiry
    underlying_price_side : PriceSide, optional
        Quoted underlying's price side to use for pricing Analysis: Bid(Bid value),
        Ask(Ask value), Mid(Mid value), Last(Last value). Optional. By default the
        "Last" price of the underlying is used.
    underlying_time_stamp : TimeStamp, optional
        Define how the timestamp of the underlying is selected:
        - Open: the opening value of the valuation_date or if not available the close of
          the previous day is used.
        - Default: the latest snapshot is used when valuation_date is today, the close
          price when valuation_date is in the past.
    volatility_model : VolatilityModel, optional
        The quantitative model used to generate the volatility surface. This may depend
        on the asset class. For Fx Volatility Surface, we currently support the SVI
        model.
    volatility_type : OptionVolatilityType, optional
        The type of volatility for the option's pricing. It can be Implied or
        SVISurface. Optional. Default to Implied. Note that if Volatility is defined,
        volatility_type is not taken into account.
    compute_payout_chart : bool, optional

    compute_volatility_payout : bool, optional

    cutoff_time : str, optional
        The cutoff time
    cutoff_time_zone : str, optional
        The cutoff time zone
    market_data_date : str, optional
        The market data date for pricing. Optional. By default, the market_data_date
        date is the valuation_date or Today
    market_value_in_deal_ccy : float, optional
        market_value_in_deal_ccy to override and that will be used as pricing analysis
        input to compute volatility_percent. Optional. No override is applied by
        default. Note that Premium takes priority over Volatility input.
    report_ccy_rate : float, optional
        Define the spot rate to use with the currency report
    risk_free_rate_percent : float, optional
        risk_free_rate_percent to override and that will be used as pricing analysis
        input to compute the option other outputs. Optional. No override is applied by
        default.
    underlying_price : float, optional
        underlying_price to override and that will be used as pricing analysis input to
        compute the option other outputs. Optional. No override is applied by default.
    valuation_date : str, optional
        The valuation date for pricing.  Optional. If not set the valuation date is
        equal to market_data_date or Today. For assets that contains a
        settlementConvention, the default valuation date  is equal to the settlementdate
        of the Asset that is usually the TradeDate+SettlementConvention.
    volatility_percent : float, optional
        volatility_percent to override and that will be used as pricing analysis input
        to compute market_value_in_deal_ccy. Optional. No override is applied by
        default. Note that if Premium is defined, volatility_percent is not taken into
        account.
    """

    def __init__(
        self,
        atm_volatility_object: Optional[BidAskMid] = None,
        butterfly10_d_object: Optional[BidAskMid] = None,
        butterfly25_d_object: Optional[BidAskMid] = None,
        domestic_deposit_rate_percent_object: Optional[BidAskMid] = None,
        foreign_deposit_rate_percent_object: Optional[BidAskMid] = None,
        forward_points_object: Optional[BidAskMid] = None,
        fx_spot_object: Optional[BidAskMid] = None,
        fx_swap_calculation_method: Optional[FxSwapCalculationMethod] = None,
        implied_volatility_object: Optional[BidAskMid] = None,
        interpolation_weight: Optional[InterpolationWeight] = None,
        option_price_side: Optional[PriceSide] = None,
        option_time_stamp: Optional[TimeStamp] = None,
        payout_custom_dates: Optional[str] = None,
        payout_scaling_interval: Optional[PayoutScaling] = None,
        price_side: Optional[PriceSide] = None,
        pricing_model_type: Optional[PricingModelType] = None,
        risk_reversal10_d_object: Optional[BidAskMid] = None,
        risk_reversal25_d_object: Optional[BidAskMid] = None,
        underlying_price_side: Optional[PriceSide] = None,
        underlying_time_stamp: Optional[TimeStamp] = None,
        volatility_model: Optional[VolatilityModel] = None,
        volatility_type: Optional[OptionVolatilityType] = None,
        compute_payout_chart: Optional[bool] = None,
        compute_volatility_payout: Optional[bool] = None,
        cutoff_time: Optional[str] = None,
        cutoff_time_zone: Optional[str] = None,
        market_data_date: Optional[str] = None,
        market_value_in_deal_ccy: Optional[float] = None,
        report_ccy_rate: Optional[float] = None,
        risk_free_rate_percent: Optional[float] = None,
        underlying_price: Optional[float] = None,
        valuation_date: Optional[str] = None,
        volatility_percent: Optional[float] = None,
    ) -> None:
        super().__init__()
        self.atm_volatility_object = atm_volatility_object
        self.butterfly10_d_object = butterfly10_d_object
        self.butterfly25_d_object = butterfly25_d_object
        self.domestic_deposit_rate_percent_object = domestic_deposit_rate_percent_object
        self.foreign_deposit_rate_percent_object = foreign_deposit_rate_percent_object
        self.forward_points_object = forward_points_object
        self.fx_spot_object = fx_spot_object
        self.fx_swap_calculation_method = fx_swap_calculation_method
        self.implied_volatility_object = implied_volatility_object
        self.interpolation_weight = interpolation_weight
        self.option_price_side = option_price_side
        self.option_time_stamp = option_time_stamp
        self.payout_custom_dates = payout_custom_dates
        self.payout_scaling_interval = payout_scaling_interval
        self.price_side = price_side
        self.pricing_model_type = pricing_model_type
        self.risk_reversal10_d_object = risk_reversal10_d_object
        self.risk_reversal25_d_object = risk_reversal25_d_object
        self.underlying_price_side = underlying_price_side
        self.underlying_time_stamp = underlying_time_stamp
        self.volatility_model = volatility_model
        self.volatility_type = volatility_type
        self.compute_payout_chart = compute_payout_chart
        self.compute_volatility_payout = compute_volatility_payout
        self.cutoff_time = cutoff_time
        self.cutoff_time_zone = cutoff_time_zone
        self.market_data_date = market_data_date
        self.market_value_in_deal_ccy = market_value_in_deal_ccy
        self.report_ccy_rate = report_ccy_rate
        self.risk_free_rate_percent = risk_free_rate_percent
        self.underlying_price = underlying_price
        self.valuation_date = valuation_date
        self.volatility_percent = volatility_percent

    @property
    def atm_volatility_object(self):
        """
        At the money volatility at Expiry
        :return: object BidAskMid
        """
        return self._get_object_parameter(BidAskMid, "atmVolatilityObject")

    @atm_volatility_object.setter
    def atm_volatility_object(self, value):
        self._set_object_parameter(BidAskMid, "atmVolatilityObject", value)

    @property
    def butterfly10_d_object(self):
        """
        BF 10 Days at Expiry
        :return: object BidAskMid
        """
        return self._get_object_parameter(BidAskMid, "butterfly10DObject")

    @butterfly10_d_object.setter
    def butterfly10_d_object(self, value):
        self._set_object_parameter(BidAskMid, "butterfly10DObject", value)

    @property
    def butterfly25_d_object(self):
        """
        BF 25 Days at Expiry
        :return: object BidAskMid
        """
        return self._get_object_parameter(BidAskMid, "butterfly25DObject")

    @butterfly25_d_object.setter
    def butterfly25_d_object(self, value):
        self._set_object_parameter(BidAskMid, "butterfly25DObject", value)

    @property
    def domestic_deposit_rate_percent_object(self):
        """
        Domestic Deposit Rate at Expiry
        :return: object BidAskMid
        """
        return self._get_object_parameter(BidAskMid, "domesticDepositRatePercentObject")

    @domestic_deposit_rate_percent_object.setter
    def domestic_deposit_rate_percent_object(self, value):
        self._set_object_parameter(BidAskMid, "domesticDepositRatePercentObject", value)

    @property
    def foreign_deposit_rate_percent_object(self):
        """
        Foreign Deposit Rate at Expiry
        :return: object BidAskMid
        """
        return self._get_object_parameter(BidAskMid, "foreignDepositRatePercentObject")

    @foreign_deposit_rate_percent_object.setter
    def foreign_deposit_rate_percent_object(self, value):
        self._set_object_parameter(BidAskMid, "foreignDepositRatePercentObject", value)

    @property
    def forward_points_object(self):
        """
        Forward Points at Expiry
        :return: object BidAskMid
        """
        return self._get_object_parameter(BidAskMid, "forwardPointsObject")

    @forward_points_object.setter
    def forward_points_object(self, value):
        self._set_object_parameter(BidAskMid, "forwardPointsObject", value)

    @property
    def fx_spot_object(self):
        """
        Spot Price
        :return: object BidAskMid
        """
        return self._get_object_parameter(BidAskMid, "fxSpotObject")

    @fx_spot_object.setter
    def fx_spot_object(self, value):
        self._set_object_parameter(BidAskMid, "fxSpotObject", value)

    @property
    def fx_swap_calculation_method(self):
        """
        The method we chose to price outrights using or not implied deposits. Possible
        values are:   FxSwap (compute outrights using swap points),
        FxSwapImpliedFromDeposit (compute swap points using deposits of currency1 and
        currency2),   DepositCcy1ImpliedFromFxSwap (compute currency1 deposits using
        swap points),   DepositCcy2ImpliedFromFxSwap (compute currency2 deposits using
        swap points).   Optional. Defaults to 'DepositCcy2ImpliedFromFxSwap'.
        :return: enum FxSwapCalculationMethod
        """
        return self._get_enum_parameter(
            FxSwapCalculationMethod, "fxSwapCalculationMethod"
        )

    @fx_swap_calculation_method.setter
    def fx_swap_calculation_method(self, value):
        self._set_enum_parameter(
            FxSwapCalculationMethod, "fxSwapCalculationMethod", value
        )

    @property
    def implied_volatility_object(self):
        """
        Implied volatility at Expiry
        :return: object BidAskMid
        """
        return self._get_object_parameter(BidAskMid, "impliedVolatilityObject")

    @implied_volatility_object.setter
    def implied_volatility_object(self, value):
        self._set_object_parameter(BidAskMid, "impliedVolatilityObject", value)

    @property
    def interpolation_weight(self):
        """
        Vol Term Structure Interpolation
        :return: object InterpolationWeight
        """
        return self._get_object_parameter(InterpolationWeight, "interpolationWeight")

    @interpolation_weight.setter
    def interpolation_weight(self, value):
        self._set_object_parameter(InterpolationWeight, "interpolationWeight", value)

    @property
    def option_price_side(self):
        """
        Quoted option's price side to use for pricing Analysis: Bid(Bid value), Ask(Ask
        value), Mid(Mid value), Last(Last value). Optional. By default the "Mid" price
        of the option is used.
        :return: enum PriceSide
        """
        return self._get_enum_parameter(PriceSide, "optionPriceSide")

    @option_price_side.setter
    def option_price_side(self, value):
        self._set_enum_parameter(PriceSide, "optionPriceSide", value)

    @property
    def option_time_stamp(self):
        """
        Define how the timestamp of the option is selected:
        - Open: the opening value of the valuation_date or if not available the close of
          the previous day is used.
        - Default: the latest snapshot is used when valuation_date is today, the close
          price when valuation_date is in the past.
        :return: enum TimeStamp
        """
        return self._get_enum_parameter(TimeStamp, "optionTimeStamp")

    @option_time_stamp.setter
    def option_time_stamp(self, value):
        self._set_enum_parameter(TimeStamp, "optionTimeStamp", value)

    @property
    def payout_custom_dates(self):
        """
        Define the custom dates for the Payout/Volatility chart
        :return: list string
        """
        return self._get_list_parameter(str, "payoutCustomDates")

    @payout_custom_dates.setter
    def payout_custom_dates(self, value):
        self._set_list_parameter(str, "payoutCustomDates", value)

    @property
    def payout_scaling_interval(self):
        """
        Define the range for the Payout/Volatility chart
        :return: object PayoutScaling
        """
        return self._get_object_parameter(PayoutScaling, "payoutScalingInterval")

    @payout_scaling_interval.setter
    def payout_scaling_interval(self, value):
        self._set_object_parameter(PayoutScaling, "payoutScalingInterval", value)

    @property
    def price_side(self):
        """
        The enumerate that specifies whether bid, ask or mid is used to price the Fx
        option. Possible values:
        - Bid
        - Ask
        - Mid
        - Last
        :return: enum PriceSide
        """
        return self._get_enum_parameter(PriceSide, "priceSide")

    @price_side.setter
    def price_side(self, value):
        self._set_enum_parameter(PriceSide, "priceSide", value)

    @property
    def pricing_model_type(self):
        """
        The enumeration that specifies the model type of pricing. Possible values:
        - BlackScholes
        - Whaley
        - Binomial
        - Trinomial
        - VannaVolga
        :return: enum PricingModelType
        """
        return self._get_enum_parameter(PricingModelType, "pricingModelType")

    @pricing_model_type.setter
    def pricing_model_type(self, value):
        self._set_enum_parameter(PricingModelType, "pricingModelType", value)

    @property
    def risk_reversal10_d_object(self):
        """
        RR 10 Days at Expiry
        :return: object BidAskMid
        """
        return self._get_object_parameter(BidAskMid, "riskReversal10DObject")

    @risk_reversal10_d_object.setter
    def risk_reversal10_d_object(self, value):
        self._set_object_parameter(BidAskMid, "riskReversal10DObject", value)

    @property
    def risk_reversal25_d_object(self):
        """
        RR 25 Days at Expiry
        :return: object BidAskMid
        """
        return self._get_object_parameter(BidAskMid, "riskReversal25DObject")

    @risk_reversal25_d_object.setter
    def risk_reversal25_d_object(self, value):
        self._set_object_parameter(BidAskMid, "riskReversal25DObject", value)

    @property
    def underlying_price_side(self):
        """
        Quoted underlying's price side to use for pricing Analysis: Bid(Bid value),
        Ask(Ask value), Mid(Mid value), Last(Last value). Optional. By default the
        "Last" price of the underlying is used.
        :return: enum PriceSide
        """
        return self._get_enum_parameter(PriceSide, "underlyingPriceSide")

    @underlying_price_side.setter
    def underlying_price_side(self, value):
        self._set_enum_parameter(PriceSide, "underlyingPriceSide", value)

    @property
    def underlying_time_stamp(self):
        """
        Define how the timestamp of the underlying is selected:
        - Open: the opening value of the valuation_date or if not available the close of
          the previous day is used.
        - Default: the latest snapshot is used when valuation_date is today, the close
          price when valuation_date is in the past.
        :return: enum TimeStamp
        """
        return self._get_enum_parameter(TimeStamp, "underlyingTimeStamp")

    @underlying_time_stamp.setter
    def underlying_time_stamp(self, value):
        self._set_enum_parameter(TimeStamp, "underlyingTimeStamp", value)

    @property
    def volatility_model(self):
        """
        The quantitative model used to generate the volatility surface. This may depend
        on the asset class. For Fx Volatility Surface, we currently support the SVI
        model.
        :return: enum VolatilityModel
        """
        return self._get_enum_parameter(VolatilityModel, "volatilityModel")

    @volatility_model.setter
    def volatility_model(self, value):
        self._set_enum_parameter(VolatilityModel, "volatilityModel", value)

    @property
    def volatility_type(self):
        """
        The type of volatility for the option's pricing. It can be Implied or
        SVISurface. Optional. Default to Implied. Note that if Volatility is defined,
        volatility_type is not taken into account.
        :return: enum OptionVolatilityType
        """
        return self._get_enum_parameter(OptionVolatilityType, "volatilityType")

    @volatility_type.setter
    def volatility_type(self, value):
        self._set_enum_parameter(OptionVolatilityType, "volatilityType", value)

    @property
    def compute_payout_chart(self):
        """
        :return: bool
        """
        return self._get_parameter("computePayoutChart")

    @compute_payout_chart.setter
    def compute_payout_chart(self, value):
        self._set_parameter("computePayoutChart", value)

    @property
    def compute_volatility_payout(self):
        """
        :return: bool
        """
        return self._get_parameter("computeVolatilityPayout")

    @compute_volatility_payout.setter
    def compute_volatility_payout(self, value):
        self._set_parameter("computeVolatilityPayout", value)

    @property
    def cutoff_time(self):
        """
        The cutoff time
        :return: str
        """
        return self._get_parameter("cutoffTime")

    @cutoff_time.setter
    def cutoff_time(self, value):
        self._set_parameter("cutoffTime", value)

    @property
    def cutoff_time_zone(self):
        """
        The cutoff time zone
        :return: str
        """
        return self._get_parameter("cutoffTimeZone")

    @cutoff_time_zone.setter
    def cutoff_time_zone(self, value):
        self._set_parameter("cutoffTimeZone", value)

    @property
    def market_data_date(self):
        """
        The market data date for pricing. Optional. By default, the market_data_date
        date is the valuation_date or Today
        :return: str
        """
        return self._get_parameter("marketDataDate")

    @market_data_date.setter
    def market_data_date(self, value):
        self._set_parameter("marketDataDate", value)

    @property
    def market_value_in_deal_ccy(self):
        """
        market_value_in_deal_ccy to override and that will be used as pricing analysis
        input to compute volatility_percent. Optional. No override is applied by
        default. Note that Premium takes priority over Volatility input.
        :return: float
        """
        return self._get_parameter("marketValueInDealCcy")

    @market_value_in_deal_ccy.setter
    def market_value_in_deal_ccy(self, value):
        self._set_parameter("marketValueInDealCcy", value)

    @property
    def report_ccy_rate(self):
        """
        Define the spot rate to use with the currency report
        :return: float
        """
        return self._get_parameter("reportCcyRate")

    @report_ccy_rate.setter
    def report_ccy_rate(self, value):
        self._set_parameter("reportCcyRate", value)

    @property
    def risk_free_rate_percent(self):
        """
        risk_free_rate_percent to override and that will be used as pricing analysis
        input to compute the option other outputs. Optional. No override is applied by
        default.
        :return: float
        """
        return self._get_parameter("riskFreeRatePercent")

    @risk_free_rate_percent.setter
    def risk_free_rate_percent(self, value):
        self._set_parameter("riskFreeRatePercent", value)

    @property
    def underlying_price(self):
        """
        underlying_price to override and that will be used as pricing analysis input to
        compute the option other outputs. Optional. No override is applied by default.
        :return: float
        """
        return self._get_parameter("underlyingPrice")

    @underlying_price.setter
    def underlying_price(self, value):
        self._set_parameter("underlyingPrice", value)

    @property
    def valuation_date(self):
        """
        The valuation date for pricing.  Optional. If not set the valuation date is
        equal to market_data_date or Today. For assets that contains a
        settlementConvention, the default valuation date  is equal to the settlementdate
        of the Asset that is usually the TradeDate+SettlementConvention.
        :return: str
        """
        return self._get_parameter("valuationDate")

    @valuation_date.setter
    def valuation_date(self, value):
        self._set_parameter("valuationDate", value)

    @property
    def volatility_percent(self):
        """
        volatility_percent to override and that will be used as pricing analysis input
        to compute market_value_in_deal_ccy. Optional. No override is applied by
        default. Note that if Premium is defined, volatility_percent is not taken into
        account.
        :return: float
        """
        return self._get_parameter("volatilityPercent")

    @volatility_percent.setter
    def volatility_percent(self, value):
        self._set_parameter("volatilityPercent", value)



import traceback
from functools import partial
from typing import Callable, Tuple, TYPE_CHECKING, Any, Dict
from typing import TypeVar, Generic

from pyee import EventEmitter

from .event import StreamEvent
from ..._core.log_reporter import _LogReporter, LogReporter
from ..._tools import DEBUG

if TYPE_CHECKING:
    from logging import Logger

T = TypeVar("T")


def __on_listener_error(
    error: Callable,
    debug: Callable,
    self,
    event: "StreamEvent",
    listener: Callable,
    exc: Exception,
) -> None:
    error(f"{self._classname} on_{event} {listener} raised exception: {exc!r}")
    debug(f"{traceback.format_exc()}")

    if DEBUG:
        raise exc


def on_listener_error(
    self, event: "StreamEvent", listener: Callable, exc: Exception
) -> None:
    __on_listener_error(self.error, self.debug, self, event, listener, exc)


def _on_listener_error(
    self, event: "StreamEvent", listener: Callable, exc: Exception
) -> None:
    __on_listener_error(self._error, self._debug, self, event, listener, exc)


def make_on_listener_error(self):
    if isinstance(self, _LogReporter):
        func = partial(_on_listener_error, self)
    elif isinstance(self, LogReporter):
        func = partial(on_listener_error, self)
    else:
        raise ValueError(f"Don't know type of self={self}")
    return func


class StreamListener(_LogReporter):
    def __init__(
        self,
        logger: "Logger",
        on_update: Callable = None,
        add_originator: bool = True,
    ) -> None:
        _LogReporter.__init__(self, logger)

        self._add_originator = add_originator
        self._emitter = EventEmitter()

        self._on_event(StreamEvent.UPDATE, on_update)

        if not hasattr(self, "_classname"):
            self._classname: str = self.__class__.__name__

        self._handler_by_event: Dict[str, Callable] = {
            StreamEvent.UPDATE: self._do_on_stream_update
        }

    def on_update(self, func: Callable) -> T:
        return self._on_event(StreamEvent.UPDATE, func, halt_if_none=True)

    def _on_event(self, event: str, func: Callable, halt_if_none: bool = False) -> T:
        if not func and halt_if_none:
            raise ValueError(f"Cannot subscribe to event {event} because no listener")

        if not func and not halt_if_none:
            return

        self._emitter.on(event, func)
        return self

    def _validate_args(self, args) -> Tuple:
        if args is None:
            raise ValueError(f"{self._classname}: args cannot be None")

        if not isinstance(args, tuple):
            args = (args,)

        return args

    def dispatch_update(self, originator, *args):
        self._debug(f"{self._classname} dispatch_update {args}")
        self._on_stream_update(originator, *args)

    def _on_stream_update(self, originator, *args) -> None:
        self._propagate_event(StreamEvent.UPDATE, originator, *args)

    def _do_on_stream_update(self, originator, *args) -> Any:
        # for override
        return args

    def _propagate_event(self, event: str, originator, *args):
        self._debug(f"{self._classname} on_{event} {args}")
        handler = self._handler_by_event[event]
        args = handler(originator, *args)
        args = self._validate_args(args)
        if self._add_originator:
            args = (self,) + args
        self._emitter.emit(event, *args)


class OMMStreamListener(StreamListener, Generic[T]):
    def __init__(
        self,
        logger: "Logger",
        on_refresh: Callable = None,
        on_status: Callable = None,
        on_update: Callable = None,
        on_complete: Callable = None,
        on_error: Callable = None,
        add_originator: bool = True,
    ) -> None:
        StreamListener.__init__(self, logger, on_update, add_originator)

        self._on_event(StreamEvent.REFRESH, on_refresh)
        self._on_event(StreamEvent.STATUS, on_status)
        self._on_event(StreamEvent.COMPLETE, on_complete)
        self._on_event(StreamEvent.ERROR, on_error)

        self._handler_by_event.update(
            {
                StreamEvent.REFRESH: self._do_on_stream_refresh,
                StreamEvent.STATUS: self._do_on_stream_status,
                StreamEvent.COMPLETE: self._do_on_stream_complete,
                StreamEvent.ERROR: self._do_on_stream_error,
            }
        )

    def on_refresh(self, func: Callable) -> T:
        return self._on_event(StreamEvent.REFRESH, func, halt_if_none=True)

    def on_status(self, func: Callable) -> T:
        return self._on_event(StreamEvent.STATUS, func, halt_if_none=True)

    def on_complete(self, func: Callable) -> T:
        return self._on_event(StreamEvent.COMPLETE, func, halt_if_none=True)

    def on_error(self, func: Callable) -> T:
        return self._on_event(StreamEvent.ERROR, func, halt_if_none=True)

    def dispatch_refresh(self, originator, *args):
        self._debug(f"{self._classname} dispatch_refresh {args}")
        self._on_stream_refresh(originator, *args)

    def _on_stream_refresh(self, originator, *args) -> None:
        self._propagate_event(StreamEvent.REFRESH, originator, *args)

    def _do_on_stream_refresh(self, originator, *args) -> Any:
        # for override
        return args

    def dispatch_status(self, originator, *args):
        self._debug(f"{self._classname} dispatch_status {args}")
        self._on_stream_status(originator, *args)

    def _on_stream_status(self, originator, *args) -> None:
        self._propagate_event(StreamEvent.STATUS, originator, *args)

    def _do_on_stream_status(self, originator, *args) -> Any:
        # for override
        return args

    def dispatch_complete(self, originator, *args):
        self._debug(f"{self._classname} dispatch_complete {args}")
        self._on_stream_complete(originator, *args)

    def _on_stream_complete(self, originator, *args) -> None:
        self._propagate_event(StreamEvent.COMPLETE, originator, *args)

    def _do_on_stream_complete(self, originator, *args) -> Any:
        # for override
        return args

    def dispatch_error(self, originator, *args):
        self._debug(f"{self._classname} dispatch_error {args}")
        self._on_stream_error(originator, *args)

    def _on_stream_error(self, originator, *args) -> None:
        self._propagate_event(StreamEvent.ERROR, originator, *args)

    def _do_on_stream_error(self, originator, *args) -> Any:
        # for override
        return args


class RDPStreamListener(StreamListener, Generic[T]):
    def __init__(
        self,
        logger: "Logger",
        on_ack: Callable = None,
        on_response: Callable = None,
        on_update: Callable = None,
        on_alarm: Callable = None,
    ) -> None:
        StreamListener.__init__(self, logger, on_update)

        self._on_event(StreamEvent.ACK, on_ack)
        self._on_event(StreamEvent.RESPONSE, on_response)
        self._on_event(StreamEvent.ALARM, on_alarm)

        self._handler_by_event.update(
            {
                StreamEvent.ACK: self._do_on_stream_ack,
                StreamEvent.RESPONSE: self._do_on_stream_response,
                StreamEvent.ALARM: self._do_on_stream_alarm,
            }
        )

    def on_ack(self, func: Callable) -> T:
        return self._on_event(StreamEvent.ACK, func, halt_if_none=True)

    def on_response(self, func: Callable) -> T:
        return self._on_event(StreamEvent.RESPONSE, func, halt_if_none=True)

    def on_alarm(self, func: Callable) -> T:
        return self._on_event(StreamEvent.ALARM, func, halt_if_none=True)

    def dispatch_ack(self, originator, *args):
        self._debug(f"{self._classname} dispatch_ack {args}")
        self._on_stream_ack(originator, *args)

    def _on_stream_ack(self, originator, *args) -> None:
        self._propagate_event(StreamEvent.ACK, originator, *args)

    def _do_on_stream_ack(self, originator, *args) -> Any:
        # for override
        return args

    def dispatch_response(self, originator, *args):
        self._debug(f"{self._classname} dispatch_response {args}")
        self._on_stream_response(originator, *args)

    def _on_stream_response(self, originator, *args) -> None:
        self._propagate_event(StreamEvent.RESPONSE, originator, *args)

    def _do_on_stream_response(self, originator, *args) -> Any:
        # for override
        return args

    def dispatch_alarm(self, originator, *args):
        self._debug(f"{self._classname} dispatch_alarm {args}")
        self._on_stream_alarm(originator, *args)

    def _on_stream_alarm(self, originator, *args) -> None:
        self._propagate_event(StreamEvent.ALARM, originator, *args)

    def _do_on_stream_alarm(self, originator, *args) -> Any:
        # for override
        return args



from ....._tools import create_repr
from ..._curves._forward_curve_request_item import ForwardCurveRequestItem
from ..._curves._forward_curve_types import (
    OptStr,
    ExtendedParams,
    CurveDefinition,
    ForwardCurveDefinitions,
    CurveParameters,
)
from ...._content_provider import ContentProviderLayer
from ...._content_type import ContentType


class Definition(ContentProviderLayer):
    """
    Parameters
    ----------
    curve_definition : SwapZcCurveDefinition, optional

    curve_parameters : SwapZcCurveParameters, optional

    forward_curve_definitions : list of ForwardCurveDefinition, optional

    curve_tag : str, optional

    extended_params : dict, optional
        If necessary other parameters.

    Methods
    -------
    get_data(session=session, on_response=on_response)
        Returns a response to the data platform
    get_data_async(session=None, on_response=None, async_mode=None)
        Returns a response asynchronously to the data platform

    Examples
    --------
     >>> import refinitiv.data.content.ipa.curves.forward_curve as forward_curve
     >>> definition = forward_curve.Definition(
     ...     curve_definition=forward_curve.SwapZcCurveDefinition(
     ...         currency="EUR",
     ...         index_name="EURIBOR",
     ...         name="EUR EURIBOR Swap ZC Curve",
     ...         discounting_tenor="OIS",
     ...     ),
     ...     forward_curve_definitions=[
     ...         forward_curve.ForwardCurveDefinition(
     ...             index_tenor="3M",
     ...             forward_curve_tag="ForwardTag",
     ...             forward_start_date="2021-02-01",
     ...             forward_curve_tenors=[
     ...                 "0D",
     ...                 "1D",
     ...                 "2D",
     ...                 "3M",
     ...                 "6M",
     ...                 "9M",
     ...                 "1Y",
     ...                 "2Y",
     ...                 "3Y",
     ...                 "4Y",
     ...                 "5Y",
     ...                 "6Y",
     ...                 "7Y",
     ...                 "8Y",
     ...                 "9Y",
     ...                 "10Y",
     ...                 "15Y",
     ...                 "20Y",
     ...                 "25Y"
     ...             ]
     ...         )
     ...     ]
     ... )
     >>> response = definition.get_data()

     Using get_data_async
     >>> import asyncio
     >>> task = definition.get_data_async()
     >>> response = asyncio.run(task)
    """

    def __init__(
        self,
        curve_definition: CurveDefinition = None,
        forward_curve_definitions: ForwardCurveDefinitions = None,
        curve_parameters: CurveParameters = None,
        curve_tag: OptStr = None,
        extended_params: ExtendedParams = None,
    ):
        request_item = ForwardCurveRequestItem(
            curve_definition=curve_definition,
            forward_curve_definitions=forward_curve_definitions,
            curve_parameters=curve_parameters,
            curve_tag=curve_tag,
        )
        super().__init__(
            content_type=ContentType.FORWARD_CURVE,
            universe=request_item,
            extended_params=extended_params,
        )

    def __repr__(self):
        return create_repr(self)



# coding: utf8

from ._repo_parameters import RepoParameters
from ..bond import PricingParameters as BondPricingParameters
from .._instrument_pricing_parameters import InstrumentPricingParameters


class UnderlyingPricingParameters(InstrumentPricingParameters):
    def __init__(
        self,
        repo_parameters=None,
        pricing_parameters_at_end=None,
        pricing_parameters_at_start=None,
        valuation_date=None,
        market_data_date=None,
    ):
        super().__init__()
        self.pricing_parameters_at_end = pricing_parameters_at_end
        self.pricing_parameters_at_start = pricing_parameters_at_start
        self.repo_parameters = repo_parameters
        self.valuation_date = valuation_date
        self.market_data_date = market_data_date

    @property
    def pricing_parameters_at_end(self):
        """
        Pricing parameters of underlying bond at Repo end date.
        :return: object BondPricingParameters
        """
        return self._get_object_parameter(
            BondPricingParameters, "pricingParametersAtEnd"
        )

    @pricing_parameters_at_end.setter
    def pricing_parameters_at_end(self, value):
        self._set_object_parameter(
            BondPricingParameters, "pricingParametersAtEnd", value
        )

    @property
    def pricing_parameters_at_start(self):
        """
        Pricing parameters of underlying bond at Repo start date.
        :return: object BondPricingParameters
        """
        return self._get_object_parameter(
            BondPricingParameters, "pricingParametersAtStart"
        )

    @pricing_parameters_at_start.setter
    def pricing_parameters_at_start(self, value):
        self._set_object_parameter(
            BondPricingParameters, "pricingParametersAtStart", value
        )

    @property
    def repo_parameters(self):
        """
        Repo parameters to be applied on underlying bond.
        :return: object RepoParameters
        """
        return self._get_object_parameter(RepoParameters, "repoParameters")

    @repo_parameters.setter
    def repo_parameters(self, value):
        self._set_object_parameter(RepoParameters, "repoParameters", value)

    @property
    def valuation_date(self):
        """
        The valuation date for pricing.
        Optional. If not set the valuation date is equal to MarketDataDate or Today. For assets that contains a settlementConvention,
        the default valuation date  is equal to the settlementdate of the Asset that is usually the TradeDate+SettlementConvention.
        :return: str
        """
        return self._get_parameter("valuationDate")

    @valuation_date.setter
    def valuation_date(self, value):
        self._set_parameter("valuationDate", value)

    @property
    def market_data_date(self):
        """
        The market data date for pricing.
        Optional. By default, the marketDataDate date is the ValuationDate or Today
        :return: str
        """
        return self._get_parameter("marketDataDate")

    @market_data_date.setter
    def market_data_date(self, value):
        self._set_parameter("marketDataDate", value)



from ._outputs import *
from ..._enums import UnderlyingType
from ..._enums import Axis
from ..._enums import DiscountingType
from ..._enums import InputVolatilityType
from ..._enums import VolatilityAdjustmentType
from ..._enums import PriceSide
from ..._enums import FxSwapCalculationMethod
from ..._enums import FxVolatilityModel
from ..._enums import TimeStamp
from ..._enums import EtiInputVolatilityType
from ..._enums import VolatilityModel
from ..._enums import Format
from ._moneyness_type import *



# coding: utf8

__all__ = ["AmortizationItem"]

from typing import Optional, Union

from .._object_definition import ObjectDefinition
from .._enums import (
    AmortizationType,
    AmortizationFrequency,
)


class AmortizationItem(ObjectDefinition):
    """
    Parameters
    ----------
    start_date : str, optional
        Start Date of an amortization section/window, or stepped rate
    end_date : str, optional
        End Date of an amortization section/window, or stepped rate
    amortization_frequency : AmortizationFrequency, optional
        Frequency of the Amortization
    amortization_type : AmortizationType or str, optional
        Amortization type Annuity, Schedule, Linear, ....
    remaining_notional : float, optional
        The Remaining Notional Amount after Amortization
    amount : float, optional
        Amortization Amount at each Amortization Date

    Examples
    ----------
    >>> amortization_item = AmortizationItem(
    ...     start_date="2021-02-11",
    ...     end_date="2022-02-11",
    ...     amount=100000,
    ...     amortization_type=AmortizationType.SCHEDULE
    ... )
    >>> amortization_item

    """

    def __init__(
        self,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
        amortization_frequency: Optional[AmortizationFrequency] = None,
        amortization_type: Union[AmortizationType, str] = None,
        remaining_notional: Optional[float] = None,
        amount: Optional[float] = None,
    ) -> None:
        super().__init__()
        self.start_date = start_date
        self.end_date = end_date
        self.amortization_frequency = amortization_frequency
        self.amortization_type = amortization_type
        self.remaining_notional = remaining_notional
        self.amount = amount

    @property
    def amortization_frequency(self):
        """
        Frequency of the Amortization
        :return: enum AmortizationFrequency
        """
        return self._get_enum_parameter(AmortizationFrequency, "amortizationFrequency")

    @amortization_frequency.setter
    def amortization_frequency(self, value):
        self._set_enum_parameter(AmortizationFrequency, "amortizationFrequency", value)

    @property
    def amortization_type(self):
        """
        Amortization type Annuity, Schedule, Linear, ....
        :return: enum AmortizationType
        """
        return self._get_enum_parameter(AmortizationType, "amortizationType")

    @amortization_type.setter
    def amortization_type(self, value):
        self._set_enum_parameter(AmortizationType, "amortizationType", value)

    @property
    def amount(self):
        """
        Amortization Amount at each Amortization Date
        :return: float
        """
        return self._get_parameter("amount")

    @amount.setter
    def amount(self, value):
        self._set_parameter("amount", value)

    @property
    def end_date(self):
        """
        End Date of an amortization section/window, or stepped rate
        :return: str
        """
        return self._get_parameter("endDate")

    @end_date.setter
    def end_date(self, value):
        self._set_parameter("endDate", value)

    @property
    def remaining_notional(self):
        """
        The Remaining Notional Amount after Amortization
        :return: float
        """
        return self._get_parameter("remainingNotional")

    @remaining_notional.setter
    def remaining_notional(self, value):
        self._set_parameter("remainingNotional", value)

    @property
    def start_date(self):
        """
        Start Date of an amortization section/window, or stepped rate
        :return: str
        """
        return self._get_parameter("startDate")

    @start_date.setter
    def start_date(self, value):
        self._set_parameter("startDate", value)



# coding: utf8

from typing import Optional

from .._instrument_definition import InstrumentDefinition
from ._enums import (
    DateRollingConvention,
    DayCountBasis,
    InterestType,
    StubRule,
    Frequency,
    AdjustInterestToPaymentDate,
    IndexCompoundingMethod,
    BusinessDayConvention,
    Direction,
)
from ._models import AmortizationItem


class BondInstrumentDefinition(InstrumentDefinition):
    """
    API endpoint for Financial Contract analytics,
    that returns calculations relevant to each contract type.

    Parameters
    ----------
    instrument_tag : str, optional
        User defined string to identify the instrument.It can be used to link output
        results to the instrument definition. Only alphabetic, numeric and '- _.#=@'
        characters are supported. Optional.
    instrument_code : str, optional
        Code to define the bond instrument. It can be an ISIN, a RIC, a CUSIP or an
        AssetId .
    end_date : str, optional
        Maturity date of the bond to override. Mandatory if instrument code has not been
        defined and is_perpetual flag has been set to false. In case an instrument code
        has been defined, value comes from bond reference data.
    direction : Direction, optional
        The direction of the leg. the possible values are:
        - 'Paid' (the cash flows of the leg are paid to the counterparty),
        - 'Received' (the cash flows of the leg are received from the counterparty).
          Optional for a single leg instrument (like a bond), in that case default value
          is Received. It is mandatory for a multi-instrument leg instrument (like Swap
          or CDS leg).
    interest_type : InterestType, optional
        A flag that indicates whether the leg is fixed or float. Possible values are:
        - 'Fixed' (the leg has a fixed coupon),
        - 'Float' (the leg has a floating rate index). Mandatory.
    notional_ccy : str, optional
        The ISO code of the notional currency. Mandatory if instrument code or
        instrument style has not been defined. In case an instrument code/style has been
        defined, value may comes from the reference data.
    notional_amount : float, optional
        The notional amount of the leg at the period start date. Optional. By default
        1,000,000 is used.
    fixed_rate_percent : float, optional
        The fixed coupon rate in percentage. It is mandatory in case of a single leg
        instrument. Otherwise, in case of multi leg instrument, it can be computed as
        the Par rate.
    spread_bp : float, optional
        The spread in basis point that is added to the floating rate index index value.
        Optional. By default 0 is used.
    interest_payment_frequency : Frequency, optional
        The frequency of the interest payments. Optional if an instrument code/style
        have been defined : in that case, value comes from reference data. Otherwise, it
        is mandatory.
    interest_calculation_method : DayCountBasis, optional
        The Day Count Basis method used to calculate the coupon interest payments.
        Mandatory.
    accrued_calculation_method : DayCountBasis, optional
        The Day Count Basis method used to calculate the accrued interest payments.
        Optional. By default, the same value than interest_calculation_method is used.
    payment_business_day_convention : BusinessDayConvention, optional
        The method to adjust dates to a working day. The possible values are:
        - ModifiedFollowing (adjusts dates according to the Modified Following
          convention - next business day unless is it goes into the next month,
          preceeding is used in that  case),
        - NextBusinessDay (adjusts dates according to the Following convention - Next
          Business Day),
        - PreviousBusinessDay (adjusts dates  according to the Preceeding convention -
          Previous Business Day),
        - NoMoving (does not adjust dates),
        - BbswModifiedFollowing (adjusts dates  according to the BBSW Modified Following
          convention). Optional. In case an instrument code/style has been defined,
          value comes from bond reference data. Otherwise 'ModifiedFollowing' is used.
    payment_roll_convention : DateRollingConvention, optional
        Method to adjust payment dates when they fall at the end of the month (28th of
        February, 30th, 31st). The possible values are:
        - Last (For setting the calculated date to the last working day),
        - Same (For setting the calculated date to the same day . In this latter case,
          the date may be moved according to the date moving convention if it is a
          non-working day),
        - Last28 (For setting the calculated date to the last working day. 28FEB being
          always considered as the last working day),
        - Same28 (For setting the calculated date to the same day .28FEB being always
          considered as the last working day). Optional. In case an instrument code has
          been defined, value comes from bond reference data. Otherwise, 'SameDay' is
          used.
    index_reset_frequency : Frequency, optional
        The reset frequency in case the leg Type is Float. Optional. By default the
        IndexTenor is used.
    index_fixing_lag : int, optional
        Defines the number of working days between the fixing date and the start of the
        coupon period ('InAdvance') or the end of the coupon period ('InArrears').
        Optional. By default 0 is used.
    first_regular_payment_date : str, optional
        The first regular coupon payment date for leg with an odd first coupon.
        Optional.
    last_regular_payment_date : str, optional
        The last regular coupon payment date for leg with an odd last coupon. Optional.
    amortization_schedule : AmortizationItem, optional
        Definition of amortizations
    payment_business_days : str, optional
        A list of coma-separated calendar codes to adjust dates (e.g. 'EMU' or 'USA').
        Optional. By default the calendar associated to notional_ccy is used.
    adjust_interest_to_payment_date : AdjustInterestToPaymentDate, optional
        A flag that indicates if the coupon dates are adjusted to the payment dates.
        Optional. By default 'false' is used.
    index_compounding_method : IndexCompoundingMethod, optional
        A flag that defines how the coupon rate is calculated from the reset floating
        rates when the reset frequency is higher than the interest payment frequency
        (e.g. daily index reset with quarterly interest payment). The possible values
        are:
        - Compounded (uses the compounded average rate from multiple fixings),
        - Average (uses the arithmetic average rate from multiple fixings),
        - Constant (uses the last published rate among multiple fixings),
        - AdjustedCompounded (uses Chinese 7-day repo fixing),
        - MexicanCompounded (uses Mexican Bremse fixing). Optional. By default
          'Constant' is used.
    interest_payment_delay : int, optional
        The number of working days between the end of coupon period and the actual
        interest payment date. Optional. By default no delay (0) is applied.
    stub_rule : StubRule, optional
        The rule that defines whether coupon roll dates are aligned on the  maturity or
        the issue date.  The possible values are:
        - ShortFirstProRata (to create a short period between the start date and the
          first coupon date, and pay a smaller amount of interest for the short
          period.All coupon dates are calculated backward from the maturity date),
        - ShortFirstFull (to create a short period between the start date and the first
          coupon date, and pay a regular coupon on the first coupon date. All coupon
          dates are calculated backward from the maturity date),
        - LongFirstFull (to create a long period between the start date and the second
          coupon date, and pay a regular coupon on the second coupon date. All coupon
          dates are calculated backward from the maturity date),
        - ShortLastProRata (to create a short period between the last payment date and
          maturity, and pay a smaller amount of interest for the short period. All
          coupon dates are calculated forward from the start date). This property may
          also be used in conjunction with first_regular_payment_date and
          last_regular_payment_date; in that case the following values can be defined:
        - Issue (all dates are aligned on the issue date),
        - Maturity (all dates are aligned on the maturity date). Optional. By default
          'Maturity' is used.
    issue_date : str, optional
        Date of issuance of the bond to override. Mandatory if instrument code has not
        been defined. In case an instrument code has been defined, value comes from bond
        reference data.
    first_accrual_date : str, optional
        Date at which bond starts accruing. Optional. In case an instrument code has
        been defined, value comes from bond reference data. Otherwise default value is
        the issue date of the bond.
    index_fixing_ric : str, optional
        The RIC that carries the fixing value. This value overrides the RIC associated
        by default with the IndexName and IndexTenor. Optional.
    is_perpetual : bool, optional
        Flag the defines wether the bond is perpetual or not in case of user defined
        bond. Optional. In case an instrument code has been defined, value comes from
        bond reference data. In case of user defined bond, default value is 'false'.
    template : str, optional
        A reference to a Adfin instrument contract or the Adfin detailed contract.
        Optional. Either instrument_code, template, or full definition must be provided.
    """

    def __init__(
        self,
        instrument_tag: Optional[str] = None,
        instrument_code: Optional[str] = None,
        end_date: Optional[str] = None,
        direction: Optional[Direction] = None,
        interest_type: Optional[InterestType] = None,
        notional_ccy: Optional[str] = None,
        notional_amount: Optional[float] = None,
        fixed_rate_percent: Optional[float] = None,
        spread_bp: Optional[float] = None,
        interest_payment_frequency: Optional[Frequency] = None,
        interest_calculation_method: Optional[DayCountBasis] = None,
        accrued_calculation_method: Optional[DayCountBasis] = None,
        payment_business_day_convention: Optional[BusinessDayConvention] = None,
        payment_roll_convention: Optional[DateRollingConvention] = None,
        index_reset_frequency: Optional[Frequency] = None,
        index_fixing_lag: Optional[int] = None,
        first_regular_payment_date: Optional[str] = None,
        last_regular_payment_date: Optional[str] = None,
        amortization_schedule: Optional[AmortizationItem] = None,
        payment_business_days: Optional[str] = None,
        adjust_interest_to_payment_date: Optional[AdjustInterestToPaymentDate] = None,
        index_compounding_method: Optional[IndexCompoundingMethod] = None,
        interest_payment_delay: Optional[int] = None,
        stub_rule: Optional[StubRule] = None,
        issue_date: Optional[str] = None,
        first_accrual_date: Optional[str] = None,
        index_fixing_ric: Optional[str] = None,
        is_perpetual: Optional[bool] = None,
        template: Optional[str] = None,
    ) -> None:
        super().__init__()
        self.instrument_tag = instrument_tag
        self.instrument_code = instrument_code
        self.end_date = end_date
        self.direction = direction
        self.interest_type = interest_type
        self.notional_ccy = notional_ccy
        self.notional_amount = notional_amount
        self.fixed_rate_percent = fixed_rate_percent
        self.spread_bp = spread_bp
        self.interest_payment_frequency = interest_payment_frequency
        self.interest_calculation_method = interest_calculation_method
        self.accrued_calculation_method = accrued_calculation_method
        self.payment_business_day_convention = payment_business_day_convention
        self.payment_roll_convention = payment_roll_convention
        self.index_reset_frequency = index_reset_frequency
        self.index_fixing_lag = index_fixing_lag
        self.first_regular_payment_date = first_regular_payment_date
        self.last_regular_payment_date = last_regular_payment_date
        self.amortization_schedule = amortization_schedule
        self.payment_business_days = payment_business_days
        self.adjust_interest_to_payment_date = adjust_interest_to_payment_date
        self.index_compounding_method = index_compounding_method
        self.interest_payment_delay = interest_payment_delay
        self.stub_rule = stub_rule
        self.issue_date = issue_date
        self.first_accrual_date = first_accrual_date
        self.index_fixing_ric = index_fixing_ric
        self.is_perpetual = is_perpetual
        self.template = template

    @classmethod
    def get_instrument_type(cls):
        return "Bond"

    @property
    def accrued_calculation_method(self):
        """
        The Day Count Basis method used to calculate the accrued interest payments.
        Optional. By default, the same value than InterestCalculationMethod is used.
        :return: enum DayCountBasis
        """
        return self._get_enum_parameter(DayCountBasis, "accruedCalculationMethod")

    @accrued_calculation_method.setter
    def accrued_calculation_method(self, value):
        self._set_enum_parameter(DayCountBasis, "accruedCalculationMethod", value)

    @property
    def adjust_interest_to_payment_date(self):
        """
        A flag that indicates if the coupon dates are adjusted to the payment dates.
        Optional. By default 'false' is used.
        :return: enum AdjustInterestToPaymentDate
        """
        return self._get_enum_parameter(
            AdjustInterestToPaymentDate, "adjustInterestToPaymentDate"
        )

    @adjust_interest_to_payment_date.setter
    def adjust_interest_to_payment_date(self, value):
        self._set_enum_parameter(
            AdjustInterestToPaymentDate, "adjustInterestToPaymentDate", value
        )

    @property
    def amortization_schedule(self):
        """
        Definition of amortizations
        :return: list AmortizationItem
        """
        return self._get_list_parameter(AmortizationItem, "amortizationSchedule")

    @amortization_schedule.setter
    def amortization_schedule(self, value):
        self._set_list_parameter(AmortizationItem, "amortizationSchedule", value)

    @property
    def direction(self):
        """
        The direction of the leg. the possible values are:
        - 'Paid' (the cash flows of the leg are paid to the counterparty),
        - 'Received' (the cash flows of the leg are received from the counterparty).
          Optional for a single leg instrument (like a bond), in that case default value
          is Received. It is mandatory for a multi-instrument leg instrument (like Swap
          or CDS leg).
        :return: enum Direction
        """
        return self._get_enum_parameter(Direction, "direction")

    @direction.setter
    def direction(self, value):
        self._set_enum_parameter(Direction, "direction", value)

    @property
    def index_compounding_method(self):
        """
        A flag that defines how the coupon rate is calculated from the reset floating
        rates when the reset frequency is higher than the interest payment frequency
        (e.g. daily index reset with quarterly interest payment). The possible values
        are:
        - Compounded (uses the compounded average rate from multiple fixings),
        - Average (uses the arithmetic average rate from multiple fixings),
        - Constant (uses the last published rate among multiple fixings),
        - AdjustedCompounded (uses Chinese 7-day repo fixing),
        - MexicanCompounded (uses Mexican Bremse fixing). Optional. By default
          'Constant' is used.
        :return: enum IndexCompoundingMethod
        """
        return self._get_enum_parameter(
            IndexCompoundingMethod, "indexCompoundingMethod"
        )

    @index_compounding_method.setter
    def index_compounding_method(self, value):
        self._set_enum_parameter(
            IndexCompoundingMethod, "indexCompoundingMethod", value
        )

    @property
    def index_reset_frequency(self):
        """
        The reset frequency in case the leg Type is Float. Optional. By default the
        IndexTenor is used.
        :return: enum Frequency
        """
        return self._get_enum_parameter(Frequency, "indexResetFrequency")

    @index_reset_frequency.setter
    def index_reset_frequency(self, value):
        self._set_enum_parameter(Frequency, "indexResetFrequency", value)

    @property
    def interest_calculation_method(self):
        """
        The Day Count Basis method used to calculate the coupon interest payments.
        Mandatory.
        :return: enum DayCountBasis
        """
        return self._get_enum_parameter(DayCountBasis, "interestCalculationMethod")

    @interest_calculation_method.setter
    def interest_calculation_method(self, value):
        self._set_enum_parameter(DayCountBasis, "interestCalculationMethod", value)

    @property
    def interest_payment_frequency(self):
        """
        The frequency of the interest payments. Optional if an instrument code/style
        have been defined : in that case, value comes from reference data. Otherwise, it
        is mandatory.
        :return: enum Frequency
        """
        return self._get_enum_parameter(Frequency, "interestPaymentFrequency")

    @interest_payment_frequency.setter
    def interest_payment_frequency(self, value):
        self._set_enum_parameter(Frequency, "interestPaymentFrequency", value)

    @property
    def interest_type(self):
        """
        A flag that indicates whether the leg is fixed or float. Possible values are:
        - 'Fixed' (the leg has a fixed coupon),
        - 'Float' (the leg has a floating rate index). Mandatory.
        :return: enum InterestType
        """
        return self._get_enum_parameter(InterestType, "interestType")

    @interest_type.setter
    def interest_type(self, value):
        self._set_enum_parameter(InterestType, "interestType", value)

    @property
    def payment_business_day_convention(self):
        """
        The method to adjust dates to a working day. The possible values are:
        - ModifiedFollowing (adjusts dates according to the Modified Following
          convention - next business day unless is it goes into the next month,
          preceeding is used in that  case),
        - NextBusinessDay (adjusts dates according to the Following convention - Next
          Business Day),
        - PreviousBusinessDay (adjusts dates  according to the Preceeding convention -
          Previous Business Day),
        - NoMoving (does not adjust dates),
        - BbswModifiedFollowing (adjusts dates  according to the BBSW Modified Following
          convention). Optional. In case an instrument code/style has been defined,
          value comes from bond reference data. Otherwise 'ModifiedFollowing' is used.
        :return: enum BusinessDayConvention
        """
        return self._get_enum_parameter(
            BusinessDayConvention, "paymentBusinessDayConvention"
        )

    @payment_business_day_convention.setter
    def payment_business_day_convention(self, value):
        self._set_enum_parameter(
            BusinessDayConvention, "paymentBusinessDayConvention", value
        )

    @property
    def payment_roll_convention(self):
        """
        Method to adjust payment dates when they fall at the end of the month (28th of
        February, 30th, 31st). The possible values are:
        - Last (For setting the calculated date to the last working day),
        - Same (For setting the calculated date to the same day . In this latter case,
          the date may be moved according to the date moving convention if it is a
          non-working day),
        - Last28 (For setting the calculated date to the last working day. 28FEB being
          always considered as the last working day),
        - Same28 (For setting the calculated date to the same day .28FEB being always
          considered as the last working day). Optional. In case an instrument code has
          been defined, value comes from bond reference data. Otherwise, 'SameDay' is
          used.
        :return: enum DateRollingConvention
        """
        return self._get_enum_parameter(DateRollingConvention, "paymentRollConvention")

    @payment_roll_convention.setter
    def payment_roll_convention(self, value):
        self._set_enum_parameter(DateRollingConvention, "paymentRollConvention", value)

    @property
    def stub_rule(self):
        """
        The rule that defines whether coupon roll dates are aligned on the  maturity or
        the issue date.  The possible values are:
        - ShortFirstProRata (to create a short period between the start date and the
          first coupon date, and pay a smaller amount of interest for the short
          period.All coupon dates are calculated backward from the maturity date),
        - ShortFirstFull (to create a short period between the start date and the first
          coupon date, and pay a regular coupon on the first coupon date. All coupon
          dates are calculated backward from the maturity date),
        - LongFirstFull (to create a long period between the start date and the second
          coupon date, and pay a regular coupon on the second coupon date. All coupon
          dates are calculated backward from the maturity date),
        - ShortLastProRata (to create a short period between the last payment date and
          maturity, and pay a smaller amount of interest for the short period. All
          coupon dates are calculated forward from the start date). This property may
          also be used in conjunction with firstRegularPaymentDate and
          lastRegularPaymentDate; in that case the following values can be defined:
        - Issue (all dates are aligned on the issue date),
        - Maturity (all dates are aligned on the maturity date). Optional. By default
          'Maturity' is used.
        :return: enum StubRule
        """
        return self._get_enum_parameter(StubRule, "stubRule")

    @stub_rule.setter
    def stub_rule(self, value):
        self._set_enum_parameter(StubRule, "stubRule", value)

    @property
    def end_date(self):
        """
        Maturity date of the bond to override. Mandatory if instrument code has not been
        defined and is_perpetual flag has been set to false. In case an instrument code
        has been defined, value comes from bond reference data.
        :return: str
        """
        return self._get_parameter("endDate")

    @end_date.setter
    def end_date(self, value):
        self._set_parameter("endDate", value)

    @property
    def first_accrual_date(self):
        """
        Date at which bond starts accruing. Optional. In case an instrument code has
        been defined, value comes from bond reference data. Otherwise default value is
        the issue date of the bond.
        :return: str
        """
        return self._get_parameter("firstAccrualDate")

    @first_accrual_date.setter
    def first_accrual_date(self, value):
        self._set_parameter("firstAccrualDate", value)

    @property
    def first_regular_payment_date(self):
        """
        The first regular coupon payment date for leg with an odd first coupon.
        Optional.
        :return: str
        """
        return self._get_parameter("firstRegularPaymentDate")

    @first_regular_payment_date.setter
    def first_regular_payment_date(self, value):
        self._set_parameter("firstRegularPaymentDate", value)

    @property
    def fixed_rate_percent(self):
        """
        The fixed coupon rate in percentage. It is mandatory in case of a single leg
        instrument. Otherwise, in case of multi leg instrument, it can be computed as
        the Par rate.
        :return: float
        """
        return self._get_parameter("fixedRatePercent")

    @fixed_rate_percent.setter
    def fixed_rate_percent(self, value):
        self._set_parameter("fixedRatePercent", value)

    @property
    def index_fixing_lag(self):
        """
        Defines the number of working days between the fixing date and the start of the
        coupon period ('InAdvance') or the end of the coupon period ('InArrears').
        Optional. By default 0 is used.
        :return: int
        """
        return self._get_parameter("indexFixingLag")

    @index_fixing_lag.setter
    def index_fixing_lag(self, value):
        self._set_parameter("indexFixingLag", value)

    @property
    def index_fixing_ric(self):
        """
        The RIC that carries the fixing value. This value overrides the RIC associated
        by default with the IndexName and IndexTenor. Optional.
        :return: str
        """
        return self._get_parameter("indexFixingRic")

    @index_fixing_ric.setter
    def index_fixing_ric(self, value):
        self._set_parameter("indexFixingRic", value)

    @property
    def instrument_code(self):
        """
        Code to define the bond instrument. It can be an ISIN, a RIC, a CUSIP or an
        AssetId .
        :return: str
        """
        return self._get_parameter("instrumentCode")

    @instrument_code.setter
    def instrument_code(self, value):
        self._set_parameter("instrumentCode", value)

    @property
    def instrument_tag(self):
        """
        User defined string to identify the instrument.It can be used to link output
        results to the instrument definition. Only alphabetic, numeric and '- _.#=@'
        characters are supported. Optional.
        :return: str
        """
        return self._get_parameter("instrumentTag")

    @instrument_tag.setter
    def instrument_tag(self, value):
        self._set_parameter("instrumentTag", value)

    @property
    def interest_payment_delay(self):
        """
        The number of working days between the end of coupon period and the actual
        interest payment date. Optional. By default no delay (0) is applied.
        :return: int
        """
        return self._get_parameter("interestPaymentDelay")

    @interest_payment_delay.setter
    def interest_payment_delay(self, value):
        self._set_parameter("interestPaymentDelay", value)

    @property
    def is_perpetual(self):
        """
        Flag the defines wether the bond is perpetual or not in case of user defined
        bond. Optional. In case an instrument code has been defined, value comes from
        bond reference data. In case of user defined bond, default value is 'false'.
        :return: bool
        """
        return self._get_parameter("isPerpetual")

    @is_perpetual.setter
    def is_perpetual(self, value):
        self._set_parameter("isPerpetual", value)

    @property
    def issue_date(self):
        """
        Date of issuance of the bond to override. Mandatory if instrument code has not
        been defined. In case an instrument code has been defined, value comes from bond
        reference data.
        :return: str
        """
        return self._get_parameter("issueDate")

    @issue_date.setter
    def issue_date(self, value):
        self._set_parameter("issueDate", value)

    @property
    def last_regular_payment_date(self):
        """
        The last regular coupon payment date for leg with an odd last coupon. Optional.
        :return: str
        """
        return self._get_parameter("lastRegularPaymentDate")

    @last_regular_payment_date.setter
    def last_regular_payment_date(self, value):
        self._set_parameter("lastRegularPaymentDate", value)

    @property
    def notional_amount(self):
        """
        The notional amount of the leg at the period start date. Optional. By default
        1,000,000 is used.
        :return: float
        """
        return self._get_parameter("notionalAmount")

    @notional_amount.setter
    def notional_amount(self, value):
        self._set_parameter("notionalAmount", value)

    @property
    def notional_ccy(self):
        """
        The ISO code of the notional currency. Mandatory if instrument code or
        instrument style has not been defined. In case an instrument code/style has been
        defined, value may comes from the reference data.
        :return: str
        """
        return self._get_parameter("notionalCcy")

    @notional_ccy.setter
    def notional_ccy(self, value):
        self._set_parameter("notionalCcy", value)

    @property
    def payment_business_days(self):
        """
        A list of coma-separated calendar codes to adjust dates (e.g. 'EMU' or 'USA').
        Optional. By default the calendar associated to notional_ccy is used.
        :return: str
        """
        return self._get_parameter("paymentBusinessDays")

    @payment_business_days.setter
    def payment_business_days(self, value):
        self._set_parameter("paymentBusinessDays", value)

    @property
    def spread_bp(self):
        """
        The spread in basis point that is added to the floating rate index index value.
        Optional. By default 0 is used.
        :return: float
        """
        return self._get_parameter("spreadBp")

    @spread_bp.setter
    def spread_bp(self, value):
        self._set_parameter("spreadBp", value)

    @property
    def template(self):
        """
        A reference to a Adfin instrument contract or the Adfin detailed contract.
        Optional. Either instrument_code, template, or full definition must be provided.
        :return: str
        """
        return self._get_parameter("template")

    @template.setter
    def template(self, value):
        self._set_parameter("template", value)



# coding: utf8

from typing import Optional

from ..._instrument_definition import ObjectDefinition


class EtiCbbcDefinition(ObjectDefinition):
    """
    Parameters
    ----------
    conversion_ratio : float, optional

    level : float, optional

    """

    def __init__(
        self,
        conversion_ratio: Optional[float] = None,
        level: Optional[float] = None,
    ) -> None:
        super().__init__()
        self.conversion_ratio = conversion_ratio
        self.level = level

    @property
    def conversion_ratio(self):
        """
        :return: float
        """
        return self._get_parameter("conversionRatio")

    @conversion_ratio.setter
    def conversion_ratio(self, value):
        self._set_parameter("conversionRatio", value)

    @property
    def level(self):
        """
        :return: float
        """
        return self._get_parameter("level")

    @level.setter
    def level(self, value):
        self._set_parameter("level", value)



from typing import Iterable, Union, TYPE_CHECKING

from ..._content_provider import ContentProviderLayer
from ..._content_type import ContentType
from ...._tools import validate_types, validate_bool_value
from .._enums import StatTypes

if TYPE_CHECKING:
    from ....content._types import ExtendedParams


class Definition(ContentProviderLayer):
    """
    This class describe parameters to retrieve holdings data breakdown by Investor Types,
    Styles, Region, Countries, Rotations and Turnovers.

    Parameters
    ----------
    universe: str, list of str
        The Universe parameter allows the user to define the companies for which the content is returned.

    stat_type: int, StatTypes
        The statType parameter specifies which statistics type to be returned.
        The types available are:
            - Investor Type (1)
            - Investment Style (2)
            - Region (3)
            - Rotation (4)
            - Country (5)
            - Metro Area (6)
            - Investor Type Parent (7)
            - Invest Style Parent (8)

    use_field_names_in_headers: bool, optional
        Return field name as column headers for data instead of title

    extended_params: ExtendedParams, optional
        If necessary other parameters.

    Examples
    --------
    >>> from refinitiv.data.content import ownership
    >>> definition = ownership.consolidated.breakdown.Definition("TRI.N", ownership.StatTypes.INVESTOR_TYPE)
    >>> response = definition.get_data()
    """

    def __init__(
        self,
        universe: Union[str, Iterable[str]],
        stat_type: Union[int, StatTypes],
        use_field_names_in_headers: bool = False,
        extended_params: "ExtendedParams" = None,
    ):
        validate_types(stat_type, [int, StatTypes], "stat_type")
        validate_bool_value(use_field_names_in_headers)

        super().__init__(
            content_type=ContentType.OWNERSHIP_CONSOLIDATED_BREAKDOWN,
            universe=universe,
            stat_type=stat_type,
            use_field_names_in_headers=use_field_names_in_headers,
            extended_params=extended_params,
        )



__all__ = (
    "AssetClass",
    "AssetState",
    "CountryCode",
    "Definition",
    "SymbolTypes",
)

from ._definition import Definition
from ._symbol_type import SymbolTypes
from ._asset_state import AssetState
from ._asset_class import AssetClass
from ._country_code import CountryCode



# coding: utf8

from enum import Enum, unique
from ...._enums._common_tools import (
    _convert_to_str,
    _normalize,
)


@unique
class DoubleBinaryType(Enum):
    DOUBLE_NO_TOUCH = "DoubleNoTouch"
    NONE = "None"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(DoubleBinaryType, _DOUBLE_BINARY_TYPE_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(
            _DOUBLE_BINARY_TYPE_VALUES_IN_LOWER_BY_DOUBLE_BINARY_TYPE, some
        )


_DOUBLE_BINARY_TYPE_VALUES = tuple(t.value for t in DoubleBinaryType)
_DOUBLE_BINARY_TYPE_VALUES_IN_LOWER_BY_DOUBLE_BINARY_TYPE = {
    name.lower(): item for name, item in DoubleBinaryType.__members__.items()
}



from typing import Optional

from .._content_type import ContentType
from ._data_provider import FilingsDataProviderLayer
from ..._tools import create_repr


class Definition(FilingsDataProviderLayer):
    """
    This class describe filename, dcn (Document Control Number), doc_id (Document ID) and
    filing_id (Financial Filing ID) to retrieve filings documents through a signed URL.

    One of the parameters is required for this class.

    Parameters
    ----------
    filename: str
        Filename is the given name of the document which also includes its file type
    dcn: str
        Document Control Number is an external identifier and an enclosed film-number specific to Edgar documents
    doc_id: str
        Document ID is a Refinitiv internal identifier assigned to financial filings documents
    filing_id: str
        Financial Filing ID is a Refinitiv internal permanent identifier assigned to each filing document

    Examples
    --------
    >>> from refinitiv.data.content import filings
    >>> definition = filings.retrieval.Definition(filename="ecpfilings_34359955599_pdf")
    >>> response = definition.get_data()
    >>> response.data.files[0].download(path="C:\\Downloads\\download_test")

    Download all files at once
    >>> response.data.files.download(path="C:\\Downloads\\download_test")
    """

    def __init__(
        self,
        filename: Optional[str] = None,
        dcn: Optional[str] = None,
        doc_id: Optional[str] = None,
        filing_id: Optional[str] = None,
    ):
        not_none_count = sum(
            param is not None for param in [filename, dcn, doc_id, filing_id]
        )
        if not_none_count == 0:
            raise ValueError(
                "One of filename, dcn, doc_id or filing_id, is required in a Definition."
            )
        elif not_none_count > 1:
            raise ValueError(
                "Only one of filename, dcn, doc_id or filing_id, can be used in a Definition"
            )

        self.filename = filename
        self.dcn = dcn
        self.doc_id = doc_id
        self.filing_id = filing_id

        super().__init__(
            content_type=ContentType.FILINGS_RETRIEVAL,
            filename=self.filename,
            dcn=self.dcn,
            doc_id=self.doc_id,
            filing_id=self.filing_id,
        )

    def __repr__(self):
        return create_repr(
            self,
            middle_path="retrieval",
            content=f"{{"
            f"filename='{self.filename}', "
            f"dcn='{self.dcn}', "
            f"doc_id='{self.doc_id}', "
            f"filing_id='{self.filing_id}'"
            f"}}",
        )



# coding: utf8

from typing import Optional

from .._base import Info
from .._enums import (
    AverageType,
    FixingFrequency,
)


class FxAverageInfo(Info):
    """
    Parameters
    ----------
    average_type : AverageType, optional
        The type of average used to compute. Possible values:
        - ArithmeticRate
        - ArithmeticStrike
        - GeometricRate
        - GeometricStrike
    fixing_frequency : FixingFrequency, optional
        The fixing's frequency. Possible values:
        - Daily
        - Weekly
        - BiWeekly
        - Monthly
        - Quaterly
        - SemiAnnual
        - Annual
    average_so_far : float, optional
        The value of the average_type
    fixing_ric_source : str, optional
        The fixing's RIC source. Default value: the first available source RIC of the Fx
        Cross Code
    fixing_start_date : str, optional
        The beginning date of the fixing period.
    include_holidays : bool, optional
        Include the holidays in the list of fixings
    include_week_ends : bool, optional
        Include the week-ends in the list of fixings
    """

    def __init__(
        self,
        average_type: Optional[AverageType] = None,
        fixing_frequency: Optional[FixingFrequency] = None,
        average_so_far: Optional[float] = None,
        fixing_ric_source: Optional[str] = None,
        fixing_start_date: Optional[str] = None,
        include_holidays: Optional[bool] = None,
        include_week_ends: Optional[bool] = None,
    ) -> None:
        super().__init__()
        self.average_type = average_type
        self.fixing_frequency = fixing_frequency
        self.average_so_far = average_so_far
        self.fixing_ric_source = fixing_ric_source
        self.fixing_start_date = fixing_start_date
        self.include_holidays = include_holidays
        self.include_week_ends = include_week_ends

    @property
    def average_type(self):
        """
        The type of average used to compute. Possible values:
        - ArithmeticRate
        - ArithmeticStrike
        - GeometricRate
        - GeometricStrike
        :return: enum AverageType
        """
        return self._get_enum_parameter(AverageType, "averageType")

    @average_type.setter
    def average_type(self, value):
        self._set_enum_parameter(AverageType, "averageType", value)

    @property
    def fixing_frequency(self):
        """
        The fixing's frequency. Possible values:
        - Daily
        - Weekly
        - BiWeekly
        - Monthly
        - Quaterly
        - SemiAnnual
        - Annual
        :return: enum FixingFrequency
        """
        return self._get_enum_parameter(FixingFrequency, "fixingFrequency")

    @fixing_frequency.setter
    def fixing_frequency(self, value):
        self._set_enum_parameter(FixingFrequency, "fixingFrequency", value)

    @property
    def average_so_far(self):
        """
        The value of the average_type
        :return: float
        """
        return self._get_parameter("averageSoFar")

    @average_so_far.setter
    def average_so_far(self, value):
        self._set_parameter("averageSoFar", value)

    @property
    def fixing_ric_source(self):
        """
        The fixing's RIC source. Default value: the first available source RIC of the Fx
        Cross Code
        :return: str
        """
        return self._get_parameter("fixingRicSource")

    @fixing_ric_source.setter
    def fixing_ric_source(self, value):
        self._set_parameter("fixingRicSource", value)

    @property
    def fixing_start_date(self):
        """
        The beginning date of the fixing period.
        :return: str
        """
        return self._get_parameter("fixingStartDate")

    @fixing_start_date.setter
    def fixing_start_date(self, value):
        self._set_parameter("fixingStartDate", value)

    @property
    def include_holidays(self):
        """
        Include the holidays in the list of fixings
        :return: bool
        """
        return self._get_parameter("includeHolidays")

    @include_holidays.setter
    def include_holidays(self, value):
        self._set_parameter("includeHolidays", value)

    @property
    def include_week_ends(self):
        """
        Include the week-ends in the list of fixings
        :return: bool
        """
        return self._get_parameter("includeWeekEnds")

    @include_week_ends.setter
    def include_week_ends(self, value):
        self._set_parameter("includeWeekEnds", value)



# coding: utf8

__all__ = ["Axis"]

from enum import Enum, unique

from ._common_tools import _convert_to_str, _normalize


@unique
class Axis(Enum):
    """
    The enumerate that specifies the unit for the axis.
    """

    X = "X"
    Y = "Y"
    Z = "Z"
    DATE = "Date"
    DELTA = "Delta"
    EXPIRY = "Expiry"
    MONEYNESS = "Moneyness"
    STRIKE = "Strike"
    TENOR = "Tenor"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(Axis, _AXIS_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_AXIS_VALUES_IN_LOWER_BY_AXIS, some)


_AXIS_VALUES = tuple(t.value for t in Axis)
_AXIS_VALUES_IN_LOWER_BY_AXIS = {
    name.lower(): item for name, item in Axis.__members__.items()
}



# coding: utf8

from typing import Optional
from ...._tools import create_repr

from ._zc_curve_types import Steps, Turns
from .._object_definition import ObjectDefinition
from ._enums import (
    CalendarAdjustment,
    CompoundingType,
    ExtrapolationMode,
    DayCountBasis,
    MarketDataAccessDeniedFallback,
    SwapPriceSide,
    ZcInterpolationMode,
)
from ._models import Step, Turn, InterestRateCurveParameters, ConvexityAdjustment
from ..._types import OptBool, OptStr, Strings


class ZcCurveParameters(ObjectDefinition):
    """
    Parameters
    ----------
    interest_calculation_method : DayCountBasis, optional
        Day count basis of the calculated zero coupon rates
    calendar_adjustment : CalendarAdjustment, optional
        Cash flow adjustment according to a calendar.
        - No: for analytic pricing (i.e. from the bond structure)
        - Null: for cash flow pricing using the calendar null
        - Weekend: for cash flow pricing using the calendar weekend
        - Calendar: for cash flow pricing using the calendar defined
                    by the parameter calendars.
    calendars : string, optional
        A list of one or more calendar codes used to define non-working days and to
        adjust coupon dates and values.
    compounding_type : CompoundingType, optional
        Output rates yield type. Values can be:
        - Continuous: continuous rate (default value)
        - MoneyMarket: money market rate
        - Compounded: compounded rate
        - Discounted: discounted rate
    convexity_adjustment : ConvexityAdjustment, optional

    extrapolation_mode : ExtrapolationMode, optional
        Extrapolation method for the curve
        - None: no extrapolation
        - Constant: constant extrapolation
        - Linear: linear extrapolation
    interpolation_mode : ZcInterpolationMode, optional
        Interpolation method for the curve. Available values are:
        - CubicDiscount: local cubic interpolation of discount factors
        - CubicRate: local cubic interpolation of rates
        - CubicSpline: a natural cubic spline
        - ForwardMonotoneConvex: forward mMonotone Convexc interpolation
        - Linear: linear interpolation
        - Log: log-linear interpolation
        - Hermite: hermite (bessel) interpolation
        - AkimaMethod: the Akima method
            (a smoother variant of local cubic interpolation)
        - FritschButlandMethod: the Fritsch-Butland method (a monotonic cubic variant)
        - KrugerMethod: the Kruger method (a monotonic cubic variant)
        - MonotonicCubicNaturalSpline: a monotonic natural cubic spline
        - MonotonicHermiteCubic: monotonic hermite (Bessel) cubic interpolation
        - TensionSpline: a tension spline
    market_data_access_denied_fallback : MarketDataAccessDeniedFallback, optional
        - ReturnError: dont price the surface and return an error (Default value)
        - IgnoreConstituents: price the surface without the error market data
        - UseDelayedData: use delayed Market Data if possible
    pivot_curve_parameters : InterestRateCurveParameters, optional

    price_side : SwapPriceSide, optional
        Price side of the instrument to be used. default value is: mid
    reference_curve_parameters : InterestRateCurveParameters, optional

    steps : list of Step, optional

    turns : list of Turn, optional
        Used to include end period rates/turns when calculating swap rate surfaces
    ignore_existing_definition : bool, optional
        The curve definition is provided in the request, so ignore the one from data
        base
    reference_tenor : str, optional
        Root tenor(s) for the xIbor dependencies
    use_convexity_adjustment : bool, optional

    use_multi_dimensional_solver : bool, optional
        Specifies the use of the multi-dimensional solver for yield curve bootstrapping.
        This solving method is required because the bootstrapping method sometimes
        creates a ZC curve which does not accurately reprice the input instruments used
        to build it. The multi-dimensional solver is recommended when cubic
        interpolation methods are used in building the curve (in other cases,
        performance might be inferior to the regular bootstrapping method). When use for
        Credit Curve it is only used when the calibrationModel is set to Bootstrap.
        - true: to use multi-dimensional solver for yield curve bootstrapping
        - false: not to use multi-dimensional solver for yield curve bootstrapping
    use_steps : bool, optional

    valuation_date : str, optional
        The valuation date. The default value is the current date.
    """

    def __init__(
        self,
        interest_calculation_method: Optional[DayCountBasis] = None,
        calendar_adjustment: Optional[CalendarAdjustment] = None,
        calendars: Strings = None,
        compounding_type: Optional[CompoundingType] = None,
        convexity_adjustment: Optional[ConvexityAdjustment] = None,
        extrapolation_mode: Optional[ExtrapolationMode] = None,
        interpolation_mode: Optional[ZcInterpolationMode] = None,
        market_data_access_denied_fallback: Optional[
            MarketDataAccessDeniedFallback
        ] = None,
        pivot_curve_parameters: Optional[InterestRateCurveParameters] = None,
        price_side: Optional[SwapPriceSide] = None,
        reference_curve_parameters: Optional[InterestRateCurveParameters] = None,
        steps: Steps = None,
        turns: Turns = None,
        ignore_existing_definition: OptBool = None,
        reference_tenor: OptStr = None,
        use_convexity_adjustment: OptBool = None,
        use_multi_dimensional_solver: OptBool = None,
        use_steps: OptBool = None,
        valuation_date: OptStr = None,
    ) -> None:
        super().__init__()
        self.interest_calculation_method = interest_calculation_method
        self.calendar_adjustment = calendar_adjustment
        self.calendars = calendars
        self.compounding_type = compounding_type
        self.convexity_adjustment = convexity_adjustment
        self.extrapolation_mode = extrapolation_mode
        self.interpolation_mode = interpolation_mode
        self.market_data_access_denied_fallback = market_data_access_denied_fallback
        self.pivot_curve_parameters = pivot_curve_parameters
        self.price_side = price_side
        self.reference_curve_parameters = reference_curve_parameters
        self.steps = steps
        self.turns = turns
        self.ignore_existing_definition = ignore_existing_definition
        self.reference_tenor = reference_tenor
        self.use_convexity_adjustment = use_convexity_adjustment
        self.use_multi_dimensional_solver = use_multi_dimensional_solver
        self.use_steps = use_steps
        self.valuation_date = valuation_date

    def __repr__(self):
        return create_repr(
            self,
            middle_path="curves.zc_curve",
            class_name=self.__class__.__name__,
        )

    @property
    def calendar_adjustment(self):
        """
        Cash flow adjustment according to a calendar.
        - No: for analytic pricing (i.e. from the bond structure)
        - Null: for cash flow pricing using the calendar null
        - Weekend: for cash flow pricing using the calendar weekend
        - Calendar: for cash flow pricing using the calendar defined
                    by the parameter calendars.
        :return: enum CalendarAdjustment
        """
        return self._get_enum_parameter(CalendarAdjustment, "calendarAdjustment")

    @calendar_adjustment.setter
    def calendar_adjustment(self, value):
        self._set_enum_parameter(CalendarAdjustment, "calendarAdjustment", value)

    @property
    def calendars(self):
        """
        A list of one or more calendar codes used to define non-working days and to
        adjust coupon dates and values.
        :return: list string
        """
        return self._get_list_parameter(str, "calendars")

    @calendars.setter
    def calendars(self, value):
        self._set_list_parameter(str, "calendars", value)

    @property
    def compounding_type(self):
        """
        Output rates yield type. Values can be:
        - Continuous: continuous rate (default value)
        - MoneyMarket: money market rate
        - Compounded: compounded rate
        - Discounted: discounted rate
        :return: enum CompoundingType
        """
        return self._get_enum_parameter(CompoundingType, "compoundingType")

    @compounding_type.setter
    def compounding_type(self, value):
        self._set_enum_parameter(CompoundingType, "compoundingType", value)

    @property
    def convexity_adjustment(self):
        """
        :return: object ConvexityAdjustment
        """
        return self._get_object_parameter(ConvexityAdjustment, "convexityAdjustment")

    @convexity_adjustment.setter
    def convexity_adjustment(self, value):
        self._set_object_parameter(ConvexityAdjustment, "convexityAdjustment", value)

    @property
    def extrapolation_mode(self):
        """
        Extrapolation method for the curve
        - None: no extrapolation
        - Constant: constant extrapolation
        - Linear: linear extrapolation
        :return: enum ExtrapolationMode
        """
        return self._get_enum_parameter(ExtrapolationMode, "extrapolationMode")

    @extrapolation_mode.setter
    def extrapolation_mode(self, value):
        self._set_enum_parameter(ExtrapolationMode, "extrapolationMode", value)

    @property
    def interest_calculation_method(self):
        """
        Day count basis of the calculated zero coupon rates
        :return: enum DayCountBasis
        """
        return self._get_enum_parameter(DayCountBasis, "interestCalculationMethod")

    @interest_calculation_method.setter
    def interest_calculation_method(self, value):
        self._set_enum_parameter(DayCountBasis, "interestCalculationMethod", value)

    @property
    def interpolation_mode(self):
        """
        Interpolation method for the curve.
        Available values are:
        - CubicDiscount: local cubic interpolation of discount factors
        - CubicRate: local cubic interpolation of rates
        - CubicSpline: a natural cubic spline
        - ForwardMonotoneConvex: forward mMonotone Convexc interpolation
        - Linear: linear interpolation
        - Log: log-linear interpolation
        - Hermite: Hermite (Bessel) interpolation
        - AkimaMethod: the Akima method (a smoother variant of local cubic interpolation)
        - FritschButlandMethod: the Fritsch-Butland method (a monotonic cubic variant)
        - KrugerMethod: the Kruger method (a monotonic cubic variant)
        - MonotonicCubicNaturalSpline: a monotonic natural cubic spline
        - MonotonicHermiteCubic: monotonic Hermite (Bessel) cubic interpolation
        - TensionSpline: a tension spline
        :return: enum ZcInterpolationMode
        """
        return self._get_enum_parameter(ZcInterpolationMode, "interpolationMode")

    @interpolation_mode.setter
    def interpolation_mode(self, value):
        self._set_enum_parameter(ZcInterpolationMode, "interpolationMode", value)

    @property
    def market_data_access_denied_fallback(self):
        """
        - ReturnError: dont price the surface and return an error (Default value)
        - IgnoreConstituents: price the surface without the error market data
        - UseDelayedData: use delayed Market Data if possible
        :return: enum MarketDataAccessDeniedFallback
        """
        return self._get_enum_parameter(
            MarketDataAccessDeniedFallback, "marketDataAccessDeniedFallback"
        )

    @market_data_access_denied_fallback.setter
    def market_data_access_denied_fallback(self, value):
        self._set_enum_parameter(
            MarketDataAccessDeniedFallback, "marketDataAccessDeniedFallback", value
        )

    @property
    def pivot_curve_parameters(self):
        """
        :return: object InterestRateCurveParameters
        """
        return self._get_object_parameter(
            InterestRateCurveParameters, "pivotCurveParameters"
        )

    @pivot_curve_parameters.setter
    def pivot_curve_parameters(self, value):
        self._set_object_parameter(
            InterestRateCurveParameters, "pivotCurveParameters", value
        )

    @property
    def price_side(self):
        """
        Price side of the instrument to be used. Default value is: Mid
        :return: enum SwapPriceSide
        """
        return self._get_enum_parameter(SwapPriceSide, "priceSide")

    @price_side.setter
    def price_side(self, value):
        self._set_enum_parameter(SwapPriceSide, "priceSide", value)

    @property
    def reference_curve_parameters(self):
        """
        :return: object InterestRateCurveParameters
        """
        return self._get_object_parameter(
            InterestRateCurveParameters, "referenceCurveParameters"
        )

    @reference_curve_parameters.setter
    def reference_curve_parameters(self, value):
        self._set_object_parameter(
            InterestRateCurveParameters, "referenceCurveParameters", value
        )

    @property
    def steps(self):
        """
        :return: list Step
        """
        return self._get_list_parameter(Step, "steps")

    @steps.setter
    def steps(self, value):
        self._set_list_parameter(Step, "steps", value)

    @property
    def turns(self):
        """
        Used to include end period rates/turns when calculating swap rate surfaces
        :return: list Turn
        """
        return self._get_list_parameter(Turn, "turns")

    @turns.setter
    def turns(self, value):
        self._set_list_parameter(Turn, "turns", value)

    @property
    def ignore_existing_definition(self):
        """
        The curve definition is provided in the request, so ignore the one from data
        base
        :return: bool
        """
        return self._get_parameter("ignoreExistingDefinition")

    @ignore_existing_definition.setter
    def ignore_existing_definition(self, value):
        self._set_parameter("ignoreExistingDefinition", value)

    @property
    def reference_tenor(self):
        """
        Root tenor(s) for the xIbor dependencies
        :return: str
        """
        return self._get_parameter("referenceTenor")

    @reference_tenor.setter
    def reference_tenor(self, value):
        self._set_parameter("referenceTenor", value)

    @property
    def use_convexity_adjustment(self):
        """
        :return: bool
        """
        return self._get_parameter("useConvexityAdjustment")

    @use_convexity_adjustment.setter
    def use_convexity_adjustment(self, value):
        self._set_parameter("useConvexityAdjustment", value)

    @property
    def use_multi_dimensional_solver(self):
        """
        Specifies the use of the multi-dimensional solver for yield curve bootstrapping.
        This solving method is required because the bootstrapping method sometimes
        creates a ZC curve which does not accurately reprice the input instruments used
        to build it. The multi-dimensional solver is recommended when cubic
        interpolation methods are used in building the curve (in other cases,
        performance might be inferior to the regular bootstrapping method). When use for
        Credit Curve it is only used when the calibrationModel is set to Bootstrap.
        - true: to use multi-dimensional solver for yield curve bootstrapping
        - false: not to use multi-dimensional solver for yield curve bootstrapping
        :return: bool
        """
        return self._get_parameter("useMultiDimensionalSolver")

    @use_multi_dimensional_solver.setter
    def use_multi_dimensional_solver(self, value):
        self._set_parameter("useMultiDimensionalSolver", value)

    @property
    def use_steps(self):
        """
        :return: bool
        """
        return self._get_parameter("useSteps")

    @use_steps.setter
    def use_steps(self, value):
        self._set_parameter("useSteps", value)

    @property
    def valuation_date(self):
        """
        The valuation date. The default value is the current date.
        :return: str
        """
        return self._get_parameter("valuationDate")

    @valuation_date.setter
    def valuation_date(self, value):
        self._set_parameter("valuationDate", value)



from enum import unique, Enum


@unique
class EventCode(Enum):
    """
    Each session can report different status events during it's lifecycle.
        StreamConnecting : Denotes the connection to the stream service within the
        session is pending.
        StreamConnected : Denotes the connection to the stream service has been
        successfully established.
        StreamDisconnected : Denotes the connection to the stream service is not
        established.
        SessionAuthenticationSuccess : Denotes the session has successfully
        authenticated this client.
        SessionAuthenticationFailed : Denotes the session has failed to authenticate
        this client.
        StreamAuthenticationSuccess: Denotes the stream has successfully
        authenticated this client.
        StreamAuthenticationFailed: Denotes the stream has failed to authenticate
        this client.
        DataRequestOk : The request for content from the session data services has
        completed successfully.
        DataRequestFailed : The request for content from the session data services
        has failed.
    """

    StreamConnecting = 1
    StreamConnected = 2
    StreamDisconnected = 3
    StreamAuthenticationSuccess = 4
    StreamAuthenticationFailed = 5
    StreamReconnecting = 6

    SessionConnecting = 21
    SessionConnected = 22
    SessionDisconnected = 23
    SessionAuthenticationSuccess = 24
    SessionAuthenticationFailed = 25
    SessionReconnecting = 26

    DataRequestOk = 61
    DataRequestFailed = 62



import abc
from datetime import datetime
from typing import TYPE_CHECKING, Callable

from ._recording_control import RecordMixin
from ..delivery._stream import StreamEvent

if TYPE_CHECKING:
    from ..content._types import OptDict
    from ._ohlc_builder import Ticks_OHLCBuilder
    from ..content._streamingprices import StreamingPrice, StreamingPrices


class StreamUpdateHandler(abc.ABC):
    updates_by_stream_name: "OptDict" = None

    @abc.abstractmethod
    def start_handling(self):
        pass

    @abc.abstractmethod
    def stop_handling(self):
        pass

    @abc.abstractmethod
    def dispose(self):
        pass


class CollectUpdates_StreamUpdateHandler(StreamUpdateHandler):
    def __init__(self, stream: "StreamingPrices", recorder: "PricingRecorder") -> None:
        self.stream = stream
        self.updates_by_stream_name = {}
        self._frequency = recorder._frequency
        self._ticks_per_bar = recorder._ticks_per_bar
        self._on_data_callback = recorder._on_data
        self._recorder = recorder

    def start_handling(self):
        for universe in self.stream.universe:
            stream: "StreamingPrice" = self.stream._streaming_prices_by_name[universe]
            stream._stream.on(StreamEvent.UPDATE, self._on_update_handler)
            stream._stream.on(StreamEvent.REFRESH, self._on_update_handler)

    def stop_handling(self):
        for universe in self.stream.universe:
            stream: "StreamingPrice" = self.stream._streaming_prices_by_name[universe]
            stream._stream.off(StreamEvent.UPDATE, self._on_update_handler)
            stream._stream.off(StreamEvent.REFRESH, self._on_update_handler)

    def dispose(self):
        self.updates_by_stream_name = {}

    def _on_update_handler(self, stream: "StreamingPrice", message: dict):
        message["Timestamp"] = datetime.now()

        updates = self.updates_by_stream_name.setdefault(stream.name, [])
        updates.append(message)

        if (
            self._frequency == "tick"
            and self._ticks_per_bar == 1
            and self._on_data_callback
        ):
            self._on_data_callback(message, self._recorder)

        self._do_on_update_handler(stream, message)

    def _do_on_update_handler(self, stream: "StreamingPrice", message: dict):
        # for override
        pass


class BuildDF_StreamUpdateHandler(CollectUpdates_StreamUpdateHandler, RecordMixin):
    def __init__(
        self,
        stream: "StreamingPrices",
        ohlc_builder: "Ticks_OHLCBuilder",
        ticks_per_bar: int,
        recorder: "PricingRecorder",
        on_record_callback: Callable = None,
    ) -> None:
        super().__init__(stream, recorder)
        self.ohlc_builder = ohlc_builder
        self.ticks_per_bar = ticks_per_bar
        self.counter = 0
        self.update_handler = self
        self.on_record_callback = on_record_callback
        self._recorder = recorder

    def _do_on_update_handler(self, stream: "StreamingPrice", message: dict):
        self.counter += 1

        if self.counter == self.ticks_per_bar:
            self.record()
            self.counter = 0



from typing import Iterable, Union

from ._base_definition import BaseDefinition
from .._content_type import ContentType
from ..._tools import validate_types, validate_bool_value


class Definition(BaseDefinition):
    """
    This class describe parameters to retrieve ESG full scores data.

    Parameters
    ----------
    universe : str, list of str
        The Universe parameter allows the user to define the company they
        want content returned for, ESG content is delivered at the Company Level.

    start : int, optional
        Start & End parameter allows the user to request
         the number of Financial Years they would like returned.

    end : int, optional
        Start & End parameter allows the user to request
        the number of Financial Years they would like returned.

    closure : str, optional
        Specifies the parameter that will be merged with the request

    use_field_names_in_headers: bool, optional
        Return field name as column headers for data instead of title

    Examples
    --------
    >>> from refinitiv.data.content import esg
    >>> definition = esg.full_scores.Definition(universe="4295904307", start=0, end=-5)
    """

    def __init__(
        self,
        universe: Union[str, Iterable[str]],
        start: int = None,
        end: int = None,
        closure: str = None,
        use_field_names_in_headers: bool = False,
    ):
        validate_types(start, [int, type(None)], "start")
        validate_types(end, [int, type(None)], "end")
        validate_bool_value(use_field_names_in_headers)

        super().__init__(
            content_type=ContentType.ESG_FULL_SCORES,
            universe=universe,
            start=start,
            end=end,
            closure=closure,
            use_field_names_in_headers=use_field_names_in_headers,
        )

    def __repr__(self):
        get_repr = super().__repr__()
        return get_repr.replace("esg", "esg.full_scores")



# coding: utf8

from typing import Optional

from ._enums import FxCrossType
from ._fx_cross_leg_definition import LegDefinition
from .._instrument_definition import InstrumentDefinition


class FxCrossInstrumentDefinition(InstrumentDefinition):
    """
    API endpoint for Financial Contract analytics,
    that returns calculations relevant to each contract type.

    Parameters
    ----------
    instrument_tag : str, optional
        User defined string to identify the instrument.It can be used to link output
        results to the instrument definition. Only alphabetic, numeric and '- _.#=@'
        characters are supported. Optional.
    legs : list of LegDefinition, optional
        Extra parameters to describe further the contract. 1 leg is mandatory for
        Forwards and NDFs contracts. 2 legs are required for Swaps, and FwdFwdSwaps
        contracts. Optional for Spot contracts.
    fx_cross_type : FxCrossType, optional
        The type of the Fx Cross instrument :  'FxSpot', 'FxForward',
        'FxNonDeliverableForward', 'FxSwap', 'MultiLeg' or 'FxForwardForward'.
        Mandatory.
    fx_cross_code : str, optional
        The ISO code of the cross currency (e.g. 'EURCHF'). Mandatory.
    ndf_fixing_settlement_ccy : str, optional
        In case of a NDF contract, the ISO code of the settlement currency (e.g. 'EUR'
        ). Optional.
    reference_spot_rate : float, optional
        Contractual Spot Rate the counterparties agreed. It is used to compute the
        traded_cross_rate as 'reference_spot_rate + traded_swap_points /
        FxSwapPointScalingFactor'. In the case of a "FxSwap" contract, it is also used
        to compute  nearLeg.ContraAmount from nearLeg.DealAmount as
        'nearLeg.ContraAmount = nearLeg.DealAmount *  (reference_spot_rate /
        FxCrossScalingFactor)'. Optional. Default value is null. In that case
        traded_cross_rate and Leg ContraAmount may not be computed.
    traded_cross_rate : float, optional
        The contractual exchange rate agreed by the two counterparties.  It is used to
        compute the ContraAmount if the amount is not filled.  In the case of a
        'FxForward' and 'FxNonDeliverableForward' contract : ContraAmount is computed as
        'DealAmount x traded_cross_rate / FxCrossScalingFactor'. In the case of a
        'FxSwap' contract : farLeg.ContraAmount is computed as 'nearLeg.DealAmount x
        traded_cross_rate / FxCrossScalingFactor'. Optional. Default value is null. It
        means that if both ContraAmount and traded_cross_rate are sot set, market value
        cannot be computed.
    traded_swap_points : float, optional
        Contractual forward points agreed by the two counterparties. It is used to
        compute the traded_cross_rate as 'reference_spot_rate + traded_swap_points /
        FxSwapPointScalingFactor'. Optional. Default value is null. In that case
        traded_cross_rate and Leg ContraAmount may not be computed.
    """

    def __init__(
        self,
        *,
        instrument_tag: Optional[str] = None,
        legs: Optional[LegDefinition] = None,
        fx_cross_type: Optional[FxCrossType] = None,
        fx_cross_code: Optional[str] = None,
        ndf_fixing_settlement_ccy: Optional[str] = None,
        reference_spot_rate: Optional[float] = None,
        traded_cross_rate: Optional[float] = None,
        traded_swap_points: Optional[float] = None,
    ) -> None:
        super().__init__()
        self.instrument_tag = instrument_tag
        self.legs = legs
        self.fx_cross_type = fx_cross_type
        self.fx_cross_code = fx_cross_code
        self.ndf_fixing_settlement_ccy = ndf_fixing_settlement_ccy
        self.reference_spot_rate = reference_spot_rate
        self.traded_cross_rate = traded_cross_rate
        self.traded_swap_points = traded_swap_points

    @classmethod
    def get_instrument_type(cls):
        return "FxCross"

    @property
    def fx_cross_type(self):
        """
        The type of the Fx Cross instrument :  'FxSpot', 'FxForward',
        'FxNonDeliverableForward', 'FxSwap', 'MultiLeg' or 'FxForwardForward'.
        Mandatory.
        :return: enum FxCrossType
        """
        return self._get_enum_parameter(FxCrossType, "fxCrossType")

    @fx_cross_type.setter
    def fx_cross_type(self, value):
        self._set_enum_parameter(FxCrossType, "fxCrossType", value)

    @property
    def legs(self):
        """
        Extra parameters to describe further the contract. 1 leg is mandatory for
        Forwards and NDFs contracts. 2 legs are required for Swaps, and FwdFwdSwaps
        contracts. Optional for Spot contracts.
        :return: list LegDefinition
        """
        return self._get_list_parameter(LegDefinition, "legs")

    @legs.setter
    def legs(self, value):
        self._set_list_parameter(LegDefinition, "legs", value)

    @property
    def fx_cross_code(self):
        """
        The ISO code of the cross currency (e.g. 'EURCHF'). Mandatory.
        :return: str
        """
        return self._get_parameter("fxCrossCode")

    @fx_cross_code.setter
    def fx_cross_code(self, value):
        self._set_parameter("fxCrossCode", value)

    @property
    def instrument_tag(self):
        """
        User defined string to identify the instrument.It can be used to link output
        results to the instrument definition. Only alphabetic, numeric and '- _.#=@'
        characters are supported. Optional.
        :return: str
        """
        return self._get_parameter("instrumentTag")

    @instrument_tag.setter
    def instrument_tag(self, value):
        self._set_parameter("instrumentTag", value)

    @property
    def ndf_fixing_settlement_ccy(self):
        """
        In case of a NDF contract, the ISO code of the settlement currency (e.g. 'EUR'
        ). Optional.
        :return: str
        """
        return self._get_parameter("ndfFixingSettlementCcy")

    @ndf_fixing_settlement_ccy.setter
    def ndf_fixing_settlement_ccy(self, value):
        self._set_parameter("ndfFixingSettlementCcy", value)

    @property
    def reference_spot_rate(self):
        """
        Contractual Spot Rate the counterparties agreed. It is used to compute the
        traded_cross_rate as 'reference_spot_rate + traded_swap_points /
        FxSwapPointScalingFactor'. In the case of a "FxSwap" contract, it is also used
        to compute  nearLeg.ContraAmount from nearLeg.DealAmount as
        'nearLeg.ContraAmount = nearLeg.DealAmount *  (reference_spot_rate /
        FxCrossScalingFactor)'. Optional. Default value is null. In that case
        traded_cross_rate and Leg ContraAmount may not be computed.
        :return: float
        """
        return self._get_parameter("referenceSpotRate")

    @reference_spot_rate.setter
    def reference_spot_rate(self, value):
        self._set_parameter("referenceSpotRate", value)

    @property
    def traded_cross_rate(self):
        """
        The contractual exchange rate agreed by the two counterparties.  It is used to
        compute the ContraAmount if the amount is not filled.  In the case of a
        'FxForward' and 'FxNonDeliverableForward' contract : ContraAmount is computed as
        'DealAmount x traded_cross_rate / FxCrossScalingFactor'. In the case of a
        'FxSwap' contract : farLeg.ContraAmount is computed as 'nearLeg.DealAmount x
        traded_cross_rate / FxCrossScalingFactor'. Optional. Default value is null. It
        means that if both ContraAmount and traded_cross_rate are sot set, market value
        cannot be computed.
        :return: float
        """
        return self._get_parameter("tradedCrossRate")

    @traded_cross_rate.setter
    def traded_cross_rate(self, value):
        self._set_parameter("tradedCrossRate", value)

    @property
    def traded_swap_points(self):
        """
        Contractual forward points agreed by the two counterparties. It is used to
        compute the traded_cross_rate as 'reference_spot_rate + traded_swap_points /
        FxSwapPointScalingFactor'. Optional. Default value is null. In that case
        traded_cross_rate and Leg ContraAmount may not be computed.
        :return: float
        """
        return self._get_parameter("tradedSwapPoints")

    @traded_swap_points.setter
    def traded_swap_points(self, value):
        self._set_parameter("tradedSwapPoints", value)



# coding: utf8

__all__ = ["DayCountBasis"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class DayCountBasis(Enum):
    """
    The enum below values the various day count fraction methods.

    The possible values are:
        - Dcb_30E_360_ISMA - Actual number of days in the coupon period calculated on
            the basis of a year of 360 days with twelve 30-day months (regardless of
            the date of the first day or last day of the period).
        - Dcb_30_360 - Actual number of days in the coupon period calculated on the
            basis of a year of 360 days with twelve 30-day months unless:
                • the last day of the period is the 31st day of a month and the first
                    day of the period is a day other than the 30th or 31st day of a month,
                    in which case the month that includes the last day shall not be
                    considered to be shortened to a 30-day month.
                • the last day of the period is the last day of the month of February,
                    in which case the month of February shall not be considered to be
                    lengthened to a 30-day month.
        - Dcb_30_360_German - For two dates (Y1,M1,D1) and (Y2,M2,D2):
                • if D1=31 then D1=30
                • if D2=31 then D2=30
                • if D1 is the last day of February then D1=30
                • if D2 is the last day of February then D2=30
            The last day of February is February 29 in leap years and February 28 in
            non leap years.
            Then the date difference is (Y2-Y1)x360+(M2-M1)*30+(D2-D1).
        - Dcb_30_360_ISDA - For two dates (Y1,M1,D1) and (Y2,M2,D2):
                • if D1 is 31, change it to 30.
                • if D2 is 31 and D1 is 30, change D2 to 30.
            Then the date difference is (Y2-Y1)x360+(M2-M1)*30+(D2-D1).
        - Dcb_30_360_US - For two dates (Y1,M1,D1) and (Y2,M2,D2):
                • if D1=31 then D1=30
                • if D2=31 and D1=30 or 31 then D2=30
                • if D1 is the last day of February then D1=30
                • if D1 is the last day of February and D2 is the last day of February then D2=30
            The last day of February is February 29 in leap years and February 28 in
            non leap years.
            The 30/360 US rule is identical to 30/360 ISDA when the EOM (end-of-month)
            convention does not apply. This indicates whether all coupon payment dates
            fall on the last day of the month. If the investment is not EOM, it will
            always pay on the same day of the month (e.g., the 10th).
            Then the date difference is (Y2-Y1)x360+(M2-M1)*30+(D2-D1).
        - Dcb_30_365_Brazil - See details into `Developer Refinitiv Data Platform APIs <https://developers.refinitiv.com/en/api-catalog/refinitiv-data-platform/refinitiv-data-platform-apis/documentation#ipa-financial-contracts-day-conventions>`_
        - Dcb_30_365_German - Similar to 30/360 (German), except that the year basis is
            treated as 365 days.
        - Dcb_30_365_ISDA - For two dates (Y1,M1,D1) and (Y2,M2,D2):
                • if D1=31 then D1=30
                • if D2=31 and D1=30 or 31 then D2=30
            Then the date difference is (Y2-Y1)*360+(M2-M1)*30+(D2-D1).
        - Dcb_30_Actual - The day count is identical to 30/360 (US) and the year basis
            is identical to Actual/Actual.
        - Dcb_30_Actual_German - The day count is identical to 30/360 (German) and the
            year basis is similar to Actual/Actual. This method was formerly used in
            the Eurobond markets.
        - Dcb_30_Actual_ISDA - See details into `Developer Refinitiv Data Platform APIs <https://developers.refinitiv.com/en/api-catalog/refinitiv-data-platform/refinitiv-data-platform-apis/documentation#ipa-financial-contracts-day-conventions>`_
        - Dcb_ActualLeapDay_360 - The day count ignores 29th February when counting days.
            The year basis is 360 days.
        - Dcb_ActualLeapDay_365 - The day count ignores 29th February when counting days.
            The year basis is 365 days.
        - Dcb_Actual_360 - The day count is the actual number of days of the period.
            The year basis is 360.
        - Dcb_Actual_364 - A special case of Actual/Actual (ISMA) when a coupon period
            contains 91 or 182 days. Actual/364 applies for some short-term instruments.
            Day count basis = 364.
        - Dcb_Actual_365 - The day count is the actual number of days of the period.
            The year basis is 365.
        - Dcb_Actual_36525 - The day count is the actual number of days of the period.
            The year basis is 365.25.
        - Dcb_Actual_365L - The day count is the actual number of days of the period.
            The year basis is calculated in the following two rules:
                • if the coupon frequency is annual, then year basis is 366 if the 29 Feb.
                    is included in the interest period, else 365.
                • if the coupon frequency is not annual, then year basis is 366 for each
                    interest period where ending date falls in a leap year, otherwise it is 365.
        - Dcb_Actual_365P - See details into `Developer Refinitiv Data Platform APIs <https://developers.refinitiv.com/en/api-catalog/refinitiv-data-platform/refinitiv-data-platform-apis/documentation#ipa-financial-contracts-day-conventions>`_
        - Dcb_Actual_365_CanadianConvention - Follows the Canadian domestic bond market
            convention. The day count basis is computed as follows:
                • if the number of days of a period is less than the actual number of
                    days in a regular coupon period the Dcb_Actual_365 convention is used.
                • otherwise: DCB= 1 - DaysRemainingInPeriod x Frequency / 365.
        - Dcb_Actual_Actual - Similar to Actual/365, except for a period that includes
            days falling in a leap year. It is calculated by DCB = number of days in a
            leap year/366 + number of days in a non-leap year/365.
            A convention is also known as Actual/365 ISDA.
        - Dcb_Actual_Actual_AFB - The DCB is calculated by Actual days / year basis where:
                • Actual days is defined as the actual days between the start date
                    (D1.M1.Y1) and end date (D2.M2.Y2).
                • Year basis is either 365 if the calculation period does not contain
                    29th Feb, or 366 if the calculation period includes 29th Feb.
        - Dcb_Actual_Actual_ISDA - Similar to Actual/365, except for a period that
            includes days falling in a leap year. It is calculated by DCB = number of
            days in a leap year/366 + number of days in a non-leap year/365.
            A convention is also known as Actual/365 ISDA.
        - Dcb_Constant - See details into `Developer Refinitiv Data Platform APIs <https://developers.refinitiv.com/en/api-catalog/refinitiv-data-platform/refinitiv-data-platform-apis/documentation#ipa-financial-contracts-day-conventions>`_
        - Dcb_WorkingDays_252 - The day count is the actual number of business days of
            the period according to the instrument calendars. The year basis is 252.
            Commonly used in the Brazilian market.
    """

    DCB_30_E_360_ISMA = "Dcb_30E_360_ISMA"
    DCB_30_360 = "Dcb_30_360"
    DCB_30_360_GERMAN = "Dcb_30_360_German"
    DCB_30_360_ISDA = "Dcb_30_360_ISDA"
    DCB_30_360_US = "Dcb_30_360_US"
    DCB_30_365_BRAZIL = "Dcb_30_365_Brazil"
    DCB_30_365_GERMAN = "Dcb_30_365_German"
    DCB_30_365_ISDA = "Dcb_30_365_ISDA"
    DCB_30_ACTUAL = "Dcb_30_Actual"
    DCB_30_ACTUAL_GERMAN = "Dcb_30_Actual_German"
    DCB_30_ACTUAL_ISDA = "Dcb_30_Actual_ISDA"
    DCB_ACTUAL_LEAP_DAY_360 = "Dcb_ActualLeapDay_360"
    DCB_ACTUAL_LEAP_DAY_365 = "Dcb_ActualLeapDay_365"
    DCB_ACTUAL_360 = "Dcb_Actual_360"
    DCB_ACTUAL_364 = "Dcb_Actual_364"
    DCB_ACTUAL_365 = "Dcb_Actual_365"
    DCB_ACTUAL_36525 = "Dcb_Actual_36525"
    DCB_ACTUAL_365_L = "Dcb_Actual_365L"
    DCB_ACTUAL_365_P = "Dcb_Actual_365P"
    DCB_ACTUAL_365_CANADIAN_CONVENTION = "Dcb_Actual_365_CanadianConvention"
    DCB_ACTUAL_ACTUAL = "Dcb_Actual_Actual"
    DCB_ACTUAL_ACTUAL_AFB = "Dcb_Actual_Actual_AFB"
    DCB_ACTUAL_ACTUAL_ISDA = "Dcb_Actual_Actual_ISDA"
    DCB_CONSTANT = "Dcb_Constant"
    DCB_WORKING_DAYS_252 = "Dcb_WorkingDays_252"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(DayCountBasis, _DAY_COUNT_BASIS_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_DAY_COUNT_BASIS_VALUES_IN_LOWER_BY_DAY_COUNT_BASIS, some)


_DAY_COUNT_BASIS_VALUES = tuple(t.value for t in DayCountBasis)
_DAY_COUNT_BASIS_VALUES_IN_LOWER_BY_DAY_COUNT_BASIS = {
    name.lower(): item for name, item in DayCountBasis.__members__.items()
}



# coding: utf-8

from .stream_connection import StreamConnection
from .omm_stream_connection import OMMStreamConnection
from .rdp_stream_connection import RDPStreamConnection

from ._stream_type import StreamType
from .event import StreamStateEvent, StreamEvent
from ._stream_listener import OMMStreamListener
from .omm_stream import OMMStream
from ._omm_stream import _OMMStream
from .rdp_stream import RDPStream
from ._rdp_stream import _RDPStream

from ._stream_cxn_cache import stream_cxn_cache
from .stream_state import StreamState
from .stream_state_manager import StreamStateManager



# coding: utf8

import time
import asyncio
from enum import Enum, unique
from typing import Callable, List, Union, TYPE_CHECKING

from ._data_grid_type import (
    DataGridType,
    data_grid_types_arg_parser,
    data_grid_type_value_by_content_type,
)
from .._content_provider import ContentProviderLayer
from .._content_type import ContentType
from .._df_build_type import DFBuildType
from ..._core.session import get_valid_session, SessionType
from ..._tools import ArgsParser, make_convert_to_enum_arg_parser
from ..._tools import create_repr
from ...errors import RDError

if TYPE_CHECKING:
    from .._types import ExtendedParams, OptBool, OptDict
    from ..._core.session import Session


@unique
class RowHeaders(Enum):
    DATE = "date"


row_headers_enum_arg_parser = make_convert_to_enum_arg_parser(RowHeaders)


def parse_row_headers(value) -> Union[RowHeaders, List[RowHeaders]]:
    if value is None:
        return []

    value = row_headers_enum_arg_parser.parse(value)

    return value


row_headers_arg_parser = ArgsParser(parse_row_headers)


def get_content_type(session: "Session") -> ContentType:
    from ...delivery._data._data_provider_factory import get_api_config

    config = get_api_config(ContentType.DATA_GRID_RDP, session.config)
    name_platform = config.setdefault("underlying-platform", DataGridType.RDP.value)
    name_platform = data_grid_types_arg_parser.get_str(name_platform)
    content_type = data_grid_type_value_by_content_type.get(name_platform)
    return content_type


def get_layout(row_headers: List[RowHeaders], content_type: ContentType) -> dict:
    layout = None

    is_rdp = content_type is ContentType.DATA_GRID_RDP
    is_udf = content_type is ContentType.DATA_GRID_UDF

    if is_udf:
        layout = {
            "layout": {
                "columns": [{"item": "dataitem"}],
                "rows": [{"item": "instrument"}],
            }
        }
    elif is_rdp:
        layout = {"output": "Col,T|Va,Row,In|"}

    if RowHeaders.DATE in row_headers:
        if is_udf:
            layout["layout"]["rows"].append({"item": "date"})

        elif is_rdp:
            output = layout["output"]
            output = output[:-1]  # delete forward slash "|"
            output = f"{output},date|"
            layout["output"] = output

    else:
        layout = ""

    if layout is None:
        raise ValueError(
            f"Layout is None, row_headers={row_headers}, content_type={content_type}"
        )

    return layout


def get_dfbuild_type(row_headers: List[RowHeaders]) -> DFBuildType:
    dfbuild_type = DFBuildType.INDEX

    if RowHeaders.DATE in row_headers:
        dfbuild_type = DFBuildType.DATE_AS_INDEX

    return dfbuild_type


class Definition(ContentProviderLayer):
    """
    This class describe the universe (list of instruments), the fields
    (a.k.a. data items) and parameters that will be requested to the data platform

    Parameters:
    ----------
    universe : str or list of str
        The list of RICs
    fields : str or list of str
        List of fundamental field names
    parameters : dict, optional
        Global parameters for fields
    row_headers : str, list of str, list of RowHeaders enum
        When this parameter is used, the output/layout parameter will be added
        to the underlying request to DataGrid RDP or UDF
    use_field_names_in_headers : bool, optional
        If value is True we add field names in headers.
    extended_params : dict, optional
        Other parameters can be provided if necessary

    Examples
    --------
     >>> from refinitiv.data.content import fundamental_and_reference
     >>> definition = fundamental_and_reference.Definition(["IBM"], ["TR.Volume"])
     >>> definition.get_data()

     Using get_data_async
     >>> import asyncio
     >>> task = definition.get_data_async()
     >>> response = asyncio.run(task)
    """

    def __init__(
        self,
        universe: Union[str, List[str]],
        fields: Union[str, List[str]],
        parameters: "OptDict" = None,
        row_headers: Union[str, List[str], List[RowHeaders]] = None,
        use_field_names_in_headers: "OptBool" = False,
        extended_params: "ExtendedParams" = None,
    ):
        self.universe = universe
        self.fields = fields
        self.parameters = parameters
        self.use_field_names_in_headers = use_field_names_in_headers
        self.extended_params = extended_params
        self.row_headers = row_headers
        super().__init__(
            content_type=ContentType.DEFAULT,
            universe=self.universe,
            fields=self.fields,
            parameters=self.parameters,
            row_headers=self.row_headers,
            use_field_names_in_headers=self.use_field_names_in_headers,
            extended_params=self.extended_params,
        )

    def _update_content_type(self, session: "Session"):
        from .._content_type import ContentType

        content_type = get_content_type(session)

        if (
            session.type == SessionType.PLATFORM
            and content_type == ContentType.DATA_GRID_UDF
        ):
            session.debug(
                f"UDF DataGrid service cannot be used with platform sessions, RDP DataGrid will be used instead. "
                f"The \"/apis/data/datagrid/underlying-platform = '{DataGridType.UDF.value}'\" "
                f"parameter will be discarded, meaning that the regular RDP DataGrid "
                f"service will be used for Fundamental and Reference data requests."
            )

            content_type = ContentType.DATA_GRID_RDP

        self._initialize(content_type, **self._kwargs)

        row_headers = self._kwargs.get("row_headers")
        row_headers = row_headers_arg_parser.get_list(row_headers)
        layout = get_layout(row_headers, content_type)
        dfbuild_type = get_dfbuild_type(row_headers)
        self._kwargs["layout"] = layout
        self._kwargs["__dfbuild_type__"] = dfbuild_type

    @staticmethod
    def make_on_response(callback: Callable) -> Callable:
        def on_response(response, data_provider, session):
            if response and response.data:
                if response.data.raw.get("ticket"):
                    return
            callback(response, data_provider, session)
        return on_response

    @staticmethod
    def _get_duration(raw_response: dict) -> Union[int, None]:
        """
        Compute the duration to sleep before next retry to request ticket status

        :param raw_response: request's response
        :type raw_response: dict
        :return: duration if response contains "estimatedDuration", None otherwise
        """
        ticket_duration = raw_response.get("estimatedDuration")
        if ticket_duration:
            ticket_duration = int(min(ticket_duration, 15000)/1000)
            return ticket_duration
        return None

    def get_data(self, session=None, on_response=None, **kwargs):
        """
        Returns a response to the data platform

        Parameters
        ----------
        session : Session, optional
            Means default session would be used
        on_response : Callable, optional
            Callable object to process retrieved data

        Returns
        -------
        Response

        Raises
        ------
        AttributeError
            If user didn't set default session.
        """
        session = get_valid_session(session)
        self._update_content_type(session)
        if self._content_type == ContentType.DATA_GRID_UDF:
            on_response_filter = on_response and self.make_on_response(on_response)
            response = super().get_data(session, on_response_filter, **kwargs)

            while response.data:
                ticket = response.data.raw.get("ticket")
                if ticket:
                    sleep_duration = self._get_duration(response.data.raw)
                    if sleep_duration:
                        time.sleep(sleep_duration)
                    else:
                        raise RDError(-1, "Receive ticket response from DataGrid without estimatedDuration")
                    response = super().get_data(session, on_response_filter, ticket=ticket)
                else:
                    break
        else:
            response = super().get_data(session, on_response, **kwargs)
        return response

    async def get_data_async(self, session=None, on_response=None, **kwargs):
        """
        Returns a response asynchronously to the data platform

        Parameters
        ----------
        session : Session, optional
            Means default session would be used
        on_response : Callable, optional
            Callable object to process retrieved data

        Returns
        -------
        Response

        Raises
        ------
        AttributeError
            If user didn't set default session.

        """
        session = get_valid_session(session)
        self._update_content_type(session)

        if self._content_type == ContentType.DATA_GRID_UDF:
            on_response_filter = on_response and self.make_on_response(on_response)
            response = await super().get_data_async(session, on_response_filter, **kwargs)

            while response.data:
                ticket = response.data.raw.get("ticket")
                if ticket:
                    sleep_duration = self._get_duration(response.data.raw)
                    if sleep_duration:
                        await asyncio.sleep(sleep_duration)
                    else:
                        raise RDError(-1, "Receive ticket response from DataGrid without estimatedDuration")
                    response = await super().get_data_async(session, on_response_filter, ticket=ticket)
                else:
                    break

        else:
            response = await super().get_data_async(session, on_response, **kwargs)
        return response

    def __repr__(self):
        return create_repr(
            self,
            content=f"{{"
            f"universe='{self.universe}', "
            f"fields='{self.fields}', "
            f"parameters='{self.parameters}', "
            f"row_headers='{self.row_headers}'"
            f"}}",
        )
